# ovirt-imageio
# Copyright (C) 2015-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import http.client as http_client
import json
import uuid

import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server
from ovirt_imageio._internal import util
from ovirt_imageio._internal.handlers import tickets

from .. import testutil
from .. import http


@pytest.fixture(
    scope="module",
    params=["daemon", "proxy"]
)
def srv(request):
    path = "test/conf/{}.conf".format(request.param)
    cfg = config.load(path)
    s = server.Server(cfg)
    s.start()
    yield s
    s.stop()


def test_method_not_allowed(srv):
    with http.ControlClient(srv.config) as c:
        res = c.request("NO_SUCH_METHO", "/tickets/")
        assert res.status == http_client.METHOD_NOT_ALLOWED


def test_no_resource(srv):
    with http.ControlClient(srv.config) as c:
        res = c.get("/no/such/resource")
        assert res.status == 404


def test_no_method(srv):
    with http.ControlClient(srv.config) as c:
        res = c.request("FOO", "/tickets/")
        assert res.status == 405


def test_get(srv, fake_time):
    ticket = testutil.create_ticket(
        ops=["read"], sparse=False, dirty=False)
    srv.auth.add(ticket)
    fake_time.now += 200
    with http.ControlClient(srv.config) as c:
        res = c.get("/tickets/%(uuid)s" % ticket)
        assert res.status == 200
        server_ticket = json.loads(res.read())
        # The server adds an expires key
        del server_ticket["expires"]
        ticket["active"] = False
        ticket["transferred"] = 0
        ticket["idle_time"] = 200
        ticket["canceled"] = False
        ticket["connections"] = 0
        assert server_ticket == ticket


def test_get_not_found(srv):
    with http.ControlClient(srv.config) as c:
        res = c.get("/tickets/%s" % uuid.uuid4())
        assert res.status == 404


def test_put(srv, fake_time):
    ticket = testutil.create_ticket(sparse=False, dirty=False)
    body = json.dumps(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/%(uuid)s" % ticket, body)
        # Server adds expires key
        ticket["expires"] = int(util.monotonic_time()) + ticket["timeout"]
        ticket["active"] = False
        ticket["idle_time"] = 0
        ticket["canceled"] = False
        ticket["connections"] = 0
        server_ticket = srv.auth.get(ticket["uuid"]).info()
        assert res.status == 200
        assert res.getheader("content-length") == "0"
        assert server_ticket == ticket


def test_put_bad_url_value(srv, fake_time):
    ticket = testutil.create_ticket(url='http://[1.2.3.4:33')
    body = json.dumps(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/%(uuid)s" % ticket, body)
        assert res.status == 400
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


def test_general_exception(srv, monkeypatch):
    def fail(*a, **kw):
        raise Exception("SECRET")
    monkeypatch.setattr(tickets.Handler, "get", fail)
    with http.ControlClient(srv.config) as c:
        res = c.get("/tickets/%s" % uuid.uuid4())
        error = res.read()
        assert res.status == http_client.INTERNAL_SERVER_ERROR
        assert b"SECRET" not in error


def test_put_no_ticket_id(srv):
    ticket = testutil.create_ticket()
    body = json.dumps(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/", body)
        assert res.status == 400
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


def test_put_invalid_json(srv):
    ticket = testutil.create_ticket()
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/%(uuid)s" % ticket, "invalid json")
        assert res.status == 400
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


# Using "timeout" confuses pytest-timeout plugin, workaround it by using
# "-timeout".
@pytest.mark.parametrize("missing", ["-timeout", "url", "size", "ops"])
def test_put_mandatory_fields(srv, missing):
    ticket = testutil.create_ticket()
    del ticket[missing.strip("-")]
    body = json.dumps(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/%(uuid)s" % ticket, body)
        assert res.status == 400
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


def test_put_invalid_timeout(srv):
    ticket = testutil.create_ticket()
    ticket["timeout"] = "invalid"
    body = json.dumps(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/%(uuid)s" % ticket, body)
        assert res.status == 400
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


def test_put_url_type_error(srv):
    ticket = testutil.create_ticket()
    ticket["url"] = 1
    body = json.dumps(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/%(uuid)s" % ticket, body)
        assert res.status == 400
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


def test_put_url_scheme_not_supported(srv):
    ticket = testutil.create_ticket()
    ticket["url"] = "notsupported:path"
    body = json.dumps(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.put("/tickets/%(uuid)s" % ticket, body)
        assert res.status == 400
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


def test_extend(srv, fake_time):
    ticket = testutil.create_ticket(sparse=False, dirty=False)
    srv.auth.add(ticket)
    patch = {"timeout": 300}
    body = json.dumps(patch)
    fake_time.now += 240
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/%(uuid)s" % ticket, body)
        ticket["expires"] = int(fake_time.now + ticket["timeout"])
        ticket["active"] = False
        ticket["idle_time"] = 240
        ticket["canceled"] = False
        ticket["connections"] = 0
        server_ticket = srv.auth.get(ticket["uuid"]).info()
        assert res.status == 200
        assert res.getheader("content-length") == "0"
        assert server_ticket == ticket


def test_extend_negative_timeout(srv):
    ticket = testutil.create_ticket(sparse=False)
    srv.auth.add(ticket)
    patch = {"timeout": -1}
    body = json.dumps(patch)
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/%(uuid)s" % ticket, body)
        assert res.status == 400


def test_get_expired_ticket(srv, fake_time):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    # Make the ticket expire.
    fake_time.now += 500
    with http.ControlClient(srv.config) as c:
        res = c.get("/tickets/%(uuid)s" % ticket)
        assert res.status == 200


def test_extend_expired_ticket(srv, fake_time):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    # Make the ticket expire.
    fake_time.now += 500
    server_ticket = srv.auth.get(ticket["uuid"]).info()
    # Extend the expired ticket.
    body = json.dumps({"timeout": 300})
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/%(uuid)s" % ticket, body)
        assert res.status == 200
        server_ticket = srv.auth.get(ticket["uuid"]).info()
        assert server_ticket["expires"] == 800


def test_extend_no_ticket_id(srv, fake_time):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    prev_ticket = srv.auth.get(ticket["uuid"]).info()
    body = json.dumps({"timeout": 300})
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/", body)
        cur_ticket = srv.auth.get(ticket["uuid"]).info()
        assert res.status == 400
        assert cur_ticket == prev_ticket


def test_extend_invalid_json(srv, fake_time):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    prev_ticket = srv.auth.get(ticket["uuid"]).info()
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/%(uuid)s" % ticket, "{invalid}")
        cur_ticket = srv.auth.get(ticket["uuid"]).info()
        assert res.status == 400
        assert cur_ticket == prev_ticket


def test_extend_no_timeout(srv, fake_time):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    prev_ticket = srv.auth.get(ticket["uuid"]).info()
    body = json.dumps({"not-a-timeout": 300})
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/%(uuid)s" % ticket, body)
        cur_ticket = srv.auth.get(ticket["uuid"]).info()
        assert res.status == 400
        assert cur_ticket == prev_ticket


def test_extend_invalid_timeout(srv, fake_time):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    prev_ticket = srv.auth.get(ticket["uuid"]).info()
    body = json.dumps({"timeout": "invalid"})
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/%(uuid)s" % ticket, body)
        cur_ticket = srv.auth.get(ticket["uuid"]).info()
        assert res.status == 400
        assert cur_ticket == prev_ticket


def test_extend_not_found(srv):
    ticket_id = str(uuid.uuid4())
    body = json.dumps({"timeout": 300})
    with http.ControlClient(srv.config) as c:
        res = c.patch("/tickets/%s" % ticket_id, body)
        assert res.status == 404


def test_idle_time_active(srv, fake_time, tmpdir):
    filename = tmpdir.join("image")
    # Note: must be big enough so the request remain active.
    size = 1024**2 * 10
    with open(str(filename), 'wb') as image:
        image.truncate(size)
    ticket = testutil.create_ticket(
        url="file://" + str(filename), ops=["read"], size=size)
    srv.auth.add(ticket)

    # Start a download, but read only 1 byte to make sure the operation becomes
    # active but do not complete.
    with http.RemoteClient(srv.config) as c:
        res = c.get("/images/" + ticket["uuid"])
        res.read(1)

        # Active ticket idle time is always 0.
        fake_time.now += 200
        assert srv.auth.get(ticket["uuid"]).idle_time == 0


def test_idle_time_inactive(srv, fake_time):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)

    # Ticket idle time starts with ticket is added.
    assert srv.auth.get(ticket["uuid"]).idle_time == 0

    # Simulate time passing without any request.
    fake_time.now += 200
    assert srv.auth.get(ticket["uuid"]).idle_time == 200


def test_idle_time_put(srv, fake_time, tmpdir):
    image = testutil.create_tempfile(tmpdir, "image", b"a" * 8192)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)

    # Request must reset idle time.
    fake_time.now += 200
    with http.RemoteClient(srv.config) as c:
        c.put("/images/" + ticket["uuid"], "b" * 8192)
        assert srv.auth.get(ticket["uuid"]).idle_time == 0


def test_idle_time_get(srv, fake_time, tmpdir):
    image = testutil.create_tempfile(tmpdir, "image", b"a" * 8192)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)

    # Request must reset idle time.
    fake_time.now += 200
    with http.RemoteClient(srv.config) as c:
        c.get("/images/" + ticket["uuid"])
        assert srv.auth.get(ticket["uuid"]).idle_time == 0


@pytest.mark.parametrize("msg", [
    pytest.param({"op": "zero", "size": 1}, id="zero"),
    pytest.param({"op": "flush"}, id="flush"),
])
def test_idle_time_patch(srv, fake_time, tmpdir, msg):
    image = testutil.create_tempfile(tmpdir, "image", b"a" * 8192)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)

    # Request must reset idle time.
    fake_time.now += 200
    body = json.dumps(msg).encode('ascii')

    with http.RemoteClient(srv.config) as c:
        c.patch("/images/" + ticket["uuid"], body,
                headers={"content-type": "application/json"})
        assert srv.auth.get(ticket["uuid"]).idle_time == 0


def test_idle_time_options(srv, tmpdir, fake_time):
    image = testutil.create_tempfile(tmpdir, "image", size=8192)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)

    # Request must reset idle time.
    fake_time.now += 200
    with http.RemoteClient(srv.config) as c:
        c.options("/images/" + ticket["uuid"])
        assert srv.auth.get(ticket["uuid"]).idle_time == 0


def test_delete_one(srv):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.delete("/tickets/%(uuid)s" % ticket)
        assert res.status == 204
        # Note: incorrect according to RFC, but required for vdsm.
        assert res.getheader("content-length") == "0"
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])


def test_delete_one_not_found(srv):
    with http.ControlClient(srv.config) as c:
        res = c.delete("/tickets/no-such-ticket")
        assert res.status == 204
        # Note: incorrect according to RFC, but required for vdsm.
        assert res.getheader("content-length") == "0"


def test_delete_all(srv):
    # Example usage: move host to maintenance
    for i in range(5):
        ticket = testutil.create_ticket(
            url="file:///tmp/foo%s" % i)
        srv.auth.add(ticket)
    with http.ControlClient(srv.config) as c:
        res = c.delete("/tickets/")
        assert res.status == 204
        # Note: incorrect according to RFC, but required for vdsm.
        assert res.getheader("content-length") == "0"
        pytest.raises(KeyError, srv.auth.get, ticket["uuid"])



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
api - imageio public client API.
"""

import json
import logging
import os
import shutil
import tarfile
import tempfile

from contextlib import contextmanager
from urllib.parse import urlparse

from .. _internal import blkhash
from .. _internal import io
from .. _internal import qemu_img
from .. _internal import qemu_nbd
from .. _internal.backends import http, nbd
from .. _internal.handlers import checksum as _checksum
from .. _internal.nbd import UnixAddress

log = logging.getLogger("client")


def upload(filename, url, cafile, buffer_size=io.BUFFER_SIZE, secure=True,
           progress=None, proxy_url=None, max_workers=io.MAX_WORKERS,
           member=None, backing_chain=True):
    """
    Upload filename to url

    Args:
        filename (str): File name for upload
        url (str): Transfer url on the host running imageio server
            e.g. https://{imageio.server}:{port}/images/{ticket-id}.
        cafile (str): Certificate file name, for example "ca.pem"
        buffer_size (int): Buffer size in bytes for reading from storage and
            sending data over HTTP connection.
        secure (bool): True for verifying server certificate and hostname.
            Default is True.
        progress (client.ProgressBar): an object implementing
            client.ProgressBar() interface.  progress.size attribute will be
            set when upload size is known, and then progress.update() will be
            called after every write or zero operation with the number bytes
            transferred.  For backward compatibility, we still support passing
            an update callable.
        proxy_url (str): Proxy url on the host running imageio as proxy, used
            if url is not accessible.
            e.g. https://{proxy.server}:{port}/images/{ticket-id}.
        max_workers (int): Maximum number of worker threads to use.
        member (str): Upload a disk with specified name from OVA file. This is
            the name reported by "tar tf vm.ova".
        backing_chain (bool): If True (default) and the image is in qcow2
            format, upload also the backing chain. If False, upload only the
            image data, leaving unallocated areas as holes, exposing data from
            the target disk backing chain. Valid only when uploding to an empty
            snapshot.
    """
    if callable(progress):
        progress = ProgressWrapper(progress)

    # Open the destination backend to get number of workers.
    with _open_http(
            url,
            "r+",
            cafile=cafile,
            secure=secure,
            proxy_url=proxy_url) as dst:

        max_workers = min(dst.max_writers, max_workers)

        # Get image format and if member specified, its offset and size.
        image_info = info(filename, member=member)

        # Open the source backend using avialable workers + extra worker used
        # for getting image extents.
        with _open_nbd(
                filename,
                image_info["format"],
                read_only=True,
                shared=max_workers + 1,
                offset=image_info.get("member-offset"),
                size=image_info.get("member-size"),
                backing_chain=backing_chain) as src:

            # Upload the image to the server.
            io.copy(
                src,
                dst,
                max_workers=max_workers,
                buffer_size=buffer_size,
                # Since we don't know if the destination image is empty, we
                # always want to zero. This can be optimized if the caller
                # knows that the image is empty.
                zero=True,
                # When uploading without a backing chain, the destination image
                # has a backing chain. We must keep holes unallocated on the so
                # they expose data from the backing chain.
                hole=backing_chain,
                progress=progress,
                name="upload")


def download(url, filename, cafile, fmt="qcow2", incremental=False,
             buffer_size=io.BUFFER_SIZE, secure=True, progress=None,
             proxy_url=None, max_workers=io.MAX_WORKERS,
             backing_file=None, backing_format=None):
    """
    Download url to filename.

    Args:
        url (str): Transfer url on the host running imageio server
            e.g. https://{imageio.server}:{port}/images/{ticket-id}.
        filename (str): Where to store downloaded data.
        cafile (str): Certificate file name, for example "ca.pem"
        fmt (str): Download file format ("raw", "qcow2"). The default is
            "qcow2" is usually the best option, supporting sparsness regardless
            of the local file system, and incremental backups.
        incremental (bool): Download only changed blocks. Valid only during
            incremetnal backup and require format="qcow2".
        buffer_size (int): Buffer size in bytes for reading from storage and
            sending data over HTTP connection.
        secure (bool): True for verifying server certificate and hostname.
            Default is True.
        progress (client.ProgressBar): an object implementing
            client.ProgressBar() interface.  progress.size attribute will be
            set when download size is known, and then progress.update() will be
            called after every read with the number bytes transferred.
        proxy_url (str): Proxy url on the host running imageio as proxy, used
            as if url is not accessible.
            e.g. https://{proxy.server}:{port}/images/{ticket-id}.
        max_workers (int): Maximum number of worker threads to use.
        backing_file (str): Set the backing file when creating qcow2 image. The
            backing file must exist.
        backing_format (str): Set the backing file format.
    """
    if incremental and fmt != "qcow2":
        raise ValueError(
            "incremental={} is incompatible with fmt={}"
            .format(incremental, fmt))

    # Open the source backend to get number of workers and image size.
    with _open_http(
            url,
            "r",
            cafile=cafile,
            secure=secure,
            proxy_url=proxy_url) as src:

        # Create a new empty image.
        qemu_img.create(
            filename,
            fmt,
            size=src.size(),
            backing_file=backing_file,
            backing_format=backing_format,
            quiet=True)

        max_workers = min(src.max_readers, max_workers)

        # Open the destination backend.
        with _open_nbd(filename, fmt, shared=max_workers) as dst:

            # Download the image from the server to the local image.
            io.copy(
                src,
                dst,
                dirty=incremental,
                max_workers=max_workers,
                buffer_size=buffer_size,
                # When downloading with a backing file, we must zero zero
                # extents which are not holes, so they hide data from the
                # backing chain.
                zero=backing_file is not None,
                # Since we always download to new empty image, we never want to
                # zero holes.
                hole=False,
                progress=progress,
                name="download")


def info(filename, member=None):
    """
    Return image information.

    If member is specified, filename must be a tar file, and the call returns
    information about file named member inside the tar file, and the offset and
    size of the member in the tar file.
    """
    if member:
        offset, size = _find_member(filename, member)
        uri = _json_uri(filename, offset, size)
        info = qemu_img.info(uri)
        info["member-offset"] = offset
        info["member-size"] = size
        return info
    else:
        return qemu_img.info(filename)


def measure(filename, dst_fmt, member=None):
    """
    Measure required size for converting filename to dst_fmt.

    If member is specified, filename must be a tar file, and the call returns
    measurement about file named member inside the tar file.
    """
    if member:
        offset, size = _find_member(filename, member)
        uri = _json_uri(filename, offset, size)
        measure = qemu_img.measure(uri, dst_fmt)
        measure["member-offset"] = offset
        measure["member-size"] = size
        return measure
    else:
        return qemu_img.measure(filename, dst_fmt)


def checksum(filename, member=None, block_size=blkhash.BLOCK_SIZE,
             algorithm=blkhash.ALGORITHM, detect_zeroes=True):
    """
    Compute image checksum.

    Arguments:
        filename (str): File name for computing checksum.
        member (str): If specified, filename must be a tar file, and the call
            returns checksum for image named member inside the tar file.
        block_size (int): block size for computing the checksum. When comparing
            to remote server checksum, the block size must match the remote
            server block size, otherwise the checksums will not match.
        algorithm (str): must one of the algorithms supported by python. See
            python documentation for available algorithms.
        detect_zeroes (bool): Detect zeroes in data extents, speeding up
            checksum calculation of preallocated images or sparse images on
            storage that does not report sparseness information.
    """
    # Get image format and if member specified, its offset and size.
    image_info = info(filename, member=member)

    with _open_nbd(
            filename,
            image_info["format"],
            read_only=True,
            offset=image_info.get("member-offset"),
            size=image_info.get("member-size")) as backend:
        buf = bytearray(block_size)
        return _checksum.compute(
            backend, buf, algorithm=algorithm, detect_zeroes=detect_zeroes)


def extents(filename, member=None, bitmap=None):
    """
    Iterate over image extents, similiar to /extents API.

    Arguments:
        filename (str): Path to file to query.
        member (str): If specified, filename must be a tar file, and the call
            returns checksum for image named member inside the tar file.
        bitmap (str): Report dirty extents using specified bitmap. Extents are
            not reported from the backing chain.
    Yields:
        Zero or dirty extents in filename.
    """
    # Get image format and if member specified, its offset and size.
    image_info = info(filename, member=member)

    with _open_nbd(
            filename,
            image_info["format"],
            read_only=True,
            bitmap=bitmap,
            offset=image_info.get("member-offset"),
            size=image_info.get("member-size")) as backend:
        for extent in backend.extents("dirty" if bitmap else "zero"):
            yield extent


class ImageioClient:
    """
    Client for imageio server.
    """

    def __init__(self, transfer_url, cafile=None, secure=True, proxy_url=None,
                 buffer_size=io.BUFFER_SIZE):
        """
        Arguments:
            transfer_url (str): Transfer url on the host running imageio server
                e.g. https://{imageio.server}:{port}/images/{ticket-id}.
            cafile (str): Certificate file name, for example "ca.pem"
            secure (bool): True for verifying server certificate and hostname.
                Default is True.
            proxy_url (str): Proxy url on the host running imageio as proxy,
                used if transfer_url is not accessible.  e.g.
                https://{proxy.server}:{port}/images/{ticket-id}.
            buffer_size (int): Buffer size in bytes for I/O operations.
        """
        self._backend = _open_http(
            transfer_url,
            "r+",
            cafile=cafile,
            secure=secure,
            proxy_url=proxy_url)
        self._buf = bytearray(buffer_size)

    @property
    def max_writers(self):
        """
        Maxumim number of concurrent clients writing data to same resource on
        imageio server.
        """
        return self._backend.max_writers

    @property
    def max_readers(self):
        """
        Maxumim number of concurrent clients reading data from same resource on
        imageio server.
        """
        return self._backend.max_readers

    def size(self):
        """
        Return image virtual size in bytes.
        """
        return self._backend.size()

    def extents(self, context="zero"):
        """
        Send extents request and iterate over returned extents.

        Arguments:
            context (str): "zero" to get zero extents, "dirty" to get dirty
                extents. Dirty extents are available only during incremental
                backup.

        Yields:
            ZeroExtent if context="zero" or DirtyExtent if context="dirty".
        """
        for extent in self._backend.extents(context):
            yield extent

    def read_from(self, reader, offset, length):
        """
        Send a PUT request and stream length bytes from reader to offset.

        Raises if offset + length is after the end of the image.

        Arguments:
            reader (object): object implementing readinto(buf).
            offset (int): offset in the image to write to.
            length (int): number of bytes you want to send.
        """
        self._backend.seek(offset)
        self._backend.read_from(reader, length, self._buf)

    def write_to(self, writer, offset, length):
        """
        Send a GET request and stream data from server to writer.

        Raises if offset + length is after the end of the image.

        Arguments:
            writer (object): object implementing write(buf).
            offset (int): offset in the image to read from.
            length (int): number of bytes to get.
        """
        self._backend.seek(offset)
        self._backend.write_to(writer, length, self._buf)

    def read(self, offset, buffer):
        """
        Send GET request, reading bytes at offset into buf.

        Always read entire buffer. Raises if offset + len(buf) is after the end
        of the image. It is more efficient to use write_to().

        Arguments:
            offset (int): offset in the image to read from.
            buffer (object): object implementing the buffer interface
                (bytearray, mmap).
        """
        if offset + len(buffer) > self._backend.size():
            raise RuntimeError("Read out of image bounds")

        self._backend.seek(offset)
        return self._backend.readinto(buffer)

    def write(self, offset, buffer):
        """
        Send PUT request, writing buf contents at offset.

        Always write entire buffer. Raises if offset + len(buf) is after the
        end of the image. It is more efficient to use read_from().

        Arguments:
            offset (int): offset in the image to write to.
            buffer (object): object implementing the buffer interface (bytes,
                bytearray, mmap).
        """
        if offset + len(buffer) > self._backend.size():
            raise RuntimeError("Write out of image bounds")

        self._backend.seek(offset)
        self._backend.write(buffer)

    def zero(self, offset, length):
        """
        Zero length bytes at offset.
        """
        self._backend.seek(offset)
        self._backend.zero(length)

    def flush(self):
        """
        Flush image data to storage.
        """
        self._backend.flush()

    def close(self):
        """
        Close the client.
        """
        self._backend.close()

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        try:
            self.close()
        except Exception:
            # Do not hide the original error.
            if t is None:
                raise
            log.exception("Error closing client")


class ProgressWrapper:
    """
    In older versions we supported passing an update() callable instead of an
    object with update() method. Wrap the callable to make it work with current
    code.
    """
    def __init__(self, update):
        self.update = update


def _find_member(tarname, name):
    with tarfile.open(tarname) as tar:
        member = tar.getmember(name)
        return member.offset_data, member.size


def _json_uri(filename, offset, size):
    # Leave the top driver to enable format probing.
    # https://lists.nongnu.org/archive/html/qemu-discuss/2020-06/msg00094.html
    nodes = {
        "file": {
            "driver": "raw",
            "offset": offset,
            "size": size,
            "file": {
                "driver": "file",
                "filename": filename,
            }
        }
    }
    return "json:" + json.dumps(nodes)


@contextmanager
def _open_nbd(filename, fmt, read_only=False, shared=1, bitmap=None,
              offset=None, size=None, backing_chain=True):
    """
    Open nbd backend.
    """
    with _tmp_dir("imageio-") as base:
        sock = UnixAddress(os.path.join(base, "sock"))
        with qemu_nbd.run(
                filename,
                fmt,
                sock,
                read_only=read_only,
                shared=shared,
                bitmap=bitmap,
                offset=offset,
                size=size,
                backing_chain=backing_chain):
            url = urlparse(sock.url())
            mode = "r" if read_only else "r+"
            yield nbd.open(url, mode=mode, dirty=bitmap is not None)


def _open_http(transfer_url, mode, cafile=None, secure=True, proxy_url=None):
    log.debug("Trying %s", transfer_url)
    url = urlparse(transfer_url)
    try:
        return http.open(url, mode, cafile=cafile, secure=secure)
    except OSError as e:
        if proxy_url is None:
            raise

        log.debug("Cannot open %s (%s), trying %s",
                  transfer_url, e, proxy_url)
        url = urlparse(proxy_url)
        return http.open(url, mode, cafile=cafile, secure=secure)


@contextmanager
def _tmp_dir(prefix):
    path = tempfile.mkdtemp(prefix=prefix)
    try:
        yield path
    finally:
        shutil.rmtree(path)



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
import os
import subprocess

import userstorage
import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server

from .. import testutil
from .. import http
from .. import storage

BACKENDS = userstorage.load_config("storage.py").BACKENDS


@pytest.fixture(
    params=[
        BACKENDS["file-512-xfs"],
        BACKENDS["file-4k-xfs"],
    ],
    ids=str
)
def user_file(request):
    with storage.Backend(request.param) as backend:
        yield backend


@pytest.fixture(scope="module")
def srv():
    cfg = config.load(["test/conf/daemon.conf"])
    s = server.Server(cfg)
    s.start()
    yield s
    s.stop()


@pytest.fixture(params=[
    pytest.param(http.RemoteClient, id="http"),
    pytest.param(http.LocalClient, id="local"),
])
def client(srv, request):
    srv.auth.clear()
    client = request.param(srv.config)
    yield client
    client.close()


def test_unupported_method(srv, client):
    res = client.request("PUT", "/images/no-such-ticket/extents")
    assert res.status == 405


def test_no_ticket_id(srv, client):
    res = client.request("GET", "/images//extents")
    assert res.status == 400


def test_no_ticket(srv, client):
    res = client.request("GET", "/images/no-such-ticket/extents")
    assert res.status == 403


def test_ticket_expired(srv, client, fake_time):
    ticket = testutil.create_ticket(timeout=300)
    srv.auth.add(ticket)

    # Make the ticket expire
    fake_time.now += 300

    res = client.request("GET", "/images/%(uuid)s/extents" % ticket)
    assert res.status == 403


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
@pytest.mark.parametrize("path", [
    "/images/%(uuid)s/extents",
    "/images/%(uuid)s/extents?context=zero",
])
def test_file_zero(srv, client, user_file, fmt, path):
    subprocess.check_call(
        ["qemu-img", "create", "-f", fmt, user_file.path, "1g"])

    # File backend uses actual size, not vitual size.
    size = os.path.getsize(user_file.path)

    ticket = testutil.create_ticket(url="file://" + user_file.path, size=size)
    srv.auth.add(ticket)

    res = client.request("GET", path % ticket)
    data = res.read()
    assert res.status == 200

    extents = json.loads(data.decode("utf-8"))
    assert extents == [
        {"start": 0, "length": size, "zero": False, "hole": False}
    ]


def test_file_ticket_not_dirty(srv, client, tmpfile):
    with open(str(tmpfile), "wb") as f:
        f.truncate(65536)

    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile), size=65536, dirty=False)
    srv.auth.add(ticket)

    res = client.request(
        "GET", "/images/%(uuid)s/extents?context=dirty" % ticket)
    res.read()
    assert res.status == 404


def test_file_does_not_support_dirty(srv, client, tmpfile):
    with open(str(tmpfile), "wb") as f:
        f.truncate(65536)

    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile), size=65536, dirty=True)
    srv.auth.add(ticket)

    res = client.request(
        "GET", "/images/%(uuid)s/extents?context=dirty" % ticket)
    res.read()
    assert res.status == 404



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pytest

from urllib.parse import urlparse

from ovirt_imageio._internal import errors
from ovirt_imageio._internal import extent
from ovirt_imageio._internal.backends import memory


def test_debugging_interface(tmpurl):
    with memory.open(urlparse("memory:"), "r+") as m:
        assert m.readable()
        assert m.writable()
        assert not m.sparse
        assert m.name == "memory"


def test_open_read_write():
    m = memory.Backend("r+")
    assert m.readable()
    assert m.writable()
    assert not m.sparse

    data = b"data"
    m.write(data)
    assert m.tell() == len(data)

    m.zero(4)
    size = len(data) + 4
    assert m.tell() == size

    content = data + b"\0" * 4
    b = bytearray(size)
    m.seek(0)
    assert m.readinto(b) == size
    assert b == content


def test_open_readonly():
    m = memory.Backend()
    assert m.readable()
    assert not m.writable()

    with pytest.raises(IOError):
        m.write(b"data")
    with pytest.raises(IOError):
        m.zero(4)
    assert m.tell() == 0
    b = bytearray(b"before")
    assert m.readinto(b) == 0
    assert b == b"before"


def test_open_writeonly():
    m = memory.Backend("w")
    assert not m.readable()
    assert m.writable()

    data = b"data"
    m.write(data)
    assert m.tell() == len(data)
    with pytest.raises(IOError):
        m.readinto(bytearray(10))
    m.flush()


def test_invalid_mode():
    with pytest.raises(ValueError):
        memory.Backend("invalid")


def test_write_inside():
    backing = bytearray(b"\0" * 8)
    m = memory.Backend("r+", data=backing)

    m.write(b"x" * 4)
    assert m.tell() == 4
    assert m.size() == 8
    assert backing == b"x" * 4 + b"\0" * 4

    m.write(b"x" * 4)
    assert m.tell() == 8
    assert m.size() == 8
    assert backing == b"x" * 8


def test_write_at_end():
    backing = bytearray(b"\0" * 4)
    m = memory.Backend("r+", data=backing)
    m.seek(4)
    m.write(b"x" * 4)

    assert m.tell() == 8
    assert m.size() == 8
    assert backing == b"\0" * 4 + b"x" * 4


def test_write_after_end():
    backing = bytearray(b"\0" * 4)
    m = memory.Backend("r+", data=backing)
    m.seek(8)
    m.write(b"x" * 4)

    assert m.tell() == 12
    assert m.size() == 12
    assert backing == b"\0" * 8 + b"x" * 4


@pytest.mark.parametrize("sparse", [True, False])
def test_zero_middle(sparse):
    m = memory.open(urlparse("memory:"), "r+", sparse=sparse)
    assert not m.sparse
    m.write(b"xxxxxxxxxxxx")
    m.seek(4)
    n = m.zero(4)
    assert n == 4
    b = bytearray(13)
    m.seek(0)
    assert m.readinto(b) == 12
    assert b[:12] == b"xxxx\x00\x00\x00\x00xxxx"


def test_zero_at_end():
    backing = bytearray(b"x" * 4)
    m = memory.Backend("r+", data=backing)
    m.seek(4)
    m.zero(4)

    assert m.tell() == 8
    assert m.size() == 8
    assert backing == b"x" * 4 + b"\0" * 4


def test_zero_after_end():
    backing = bytearray(b"x" * 4)
    m = memory.Backend("r+", data=backing)
    m.seek(8)
    m.zero(4)

    assert m.tell() == 12
    assert m.size() == 12
    assert backing == b"x" * 4 + b"\0" * 8


def test_close():
    m = memory.Backend("r+")
    m.close()
    # All operations should fail now with:
    #     ValueError: I/O operation on closed file
    with pytest.raises(ValueError):
        m.write("more")
    with pytest.raises(ValueError):
        m.readinto(bytearray(10))


def test_context_manager():
    with memory.Backend("r+") as m:
        m.write(b"data")
    with pytest.raises(ValueError):
        m.write("more")


def test_close_error():

    def close():
        raise IOError("backend error")

    with pytest.raises(IOError):
        with memory.Backend("r+") as m:
            m.close = close


def test_propagate_user_error():

    class UserError(Exception):
        pass

    def close():
        raise IOError("backend error")

    with pytest.raises(UserError):
        with memory.Backend("r+") as m:
            m.close = close
            raise UserError("user error")


def test_dirty():
    # backend created clean
    m = memory.Backend("r+", data=bytearray(b"data"))
    assert not m.dirty

    # write ans zero dirty the backend
    m.write(b"01234")
    assert m.dirty
    m.flush()
    assert not m.dirty
    m.zero(5)
    assert m.dirty
    m.flush()
    assert not m.dirty

    # readinto, seek do not affect dirty.
    b = bytearray(10)
    m.seek(0)
    assert not m.dirty
    m.readinto(b)
    assert not m.dirty


def test_size():
    m = memory.Backend("r+", data=bytearray(b"data"))
    assert m.size() == 4
    assert m.tell() == 0
    m.zero(5)
    m.seek(3)
    assert m.size() == 5
    assert m.tell() == 3


def test_extents():
    m = memory.Backend(data=bytearray(b"data"))
    assert list(m.extents()) == [extent.ZeroExtent(0, 4, False, False)]


def test_extents_dirty():
    m = memory.Backend(data=bytearray(b"data"))
    with pytest.raises(errors.UnsupportedOperation):
        list(m.extents(context="dirty"))


def test_user_extents():
    extents = {
        "zero": [
            extent.ZeroExtent(0, 32, False, False),
            extent.ZeroExtent(0, 32, True, False),
        ],
        "dirty": [
            extent.DirtyExtent(0, 16, True, False),
            extent.DirtyExtent(0, 48, False, False),
        ]
    }
    data = b"a" * 32 + b"\0" * 32

    m = memory.Backend(data=data, extents=extents)

    assert list(m.extents()) == extents["zero"]
    assert list(m.extents("zero")) == extents["zero"]
    assert list(m.extents("dirty")) == extents["dirty"]


def test_read_from():
    size = 128
    src = memory.Backend(data=bytearray(b"x" * size))
    dst = memory.ReaderFrom("r+", bytearray(b"y" * size))
    buf = bytearray(32)
    dst.read_from(src, size, buf)

    assert src.tell() == size
    assert dst.tell() == size
    assert src.data() == dst.data()


def test_read_from_some():
    src = memory.Backend(data=bytearray(b"x" * 128))
    dst = memory.ReaderFrom("r+", data=bytearray(b"y" * 128))
    buf = bytearray(32)

    src.seek(32)
    dst.seek(32)
    dst.read_from(src, 64, buf)

    assert src.tell() == 96
    assert dst.tell() == 96
    assert dst.data() == b"y" * 32 + b"x" * 64 + b"y" * 32


def test_write_to():
    size = 128
    src = memory.WriterTo(data=bytearray(b"x" * size))
    dst = memory.Backend("r+", data=bytearray(b"y" * size))
    buf = bytearray(32)
    src.write_to(dst, size, buf)

    assert src.tell() == size
    assert dst.tell() == size
    assert src.data() == dst.data()


def test_write_to_some():
    size = 128
    src = memory.WriterTo(data=bytearray(b"x" * size))
    dst = memory.Backend("r+", data=bytearray(b"y" * size))
    buf = bytearray(32)

    src.seek(32)
    dst.seek(32)
    src.write_to(dst, 64, buf)

    assert src.tell() == 96
    assert dst.tell() == 96
    assert dst.data() == b"y" * 32 + b"x" * 64 + b"y" * 32


def test_clone():
    backing = bytearray(b"x" * 128)
    a = memory.Backend(mode="r+", data=backing)
    b = a.clone()

    # Backends are indentical on creation.
    assert a.size() == b.size()
    assert a.tell() == b.tell()
    assert a.data() == b.data()

    # Writng to a modifies the shared backing but does not modify b position.
    a.write(b"y" * 64)
    assert a.tell() == 64
    assert b.tell() == 0
    assert a.dirty
    assert not b.dirty
    assert a.data() == b.data()

    # Writng to b modifies the shared backing but does not modify a position.
    b.seek(64)
    b.write(b"y" * 64)
    assert a.tell() == 64
    assert b.tell() == 128
    assert b.dirty
    assert a.data() == b.data()

    # Extending the backing show in both backends.
    a.seek(128)
    a.zero(64)
    assert a.tell() == 192
    assert b.tell() == 128
    assert a.size() == b.size()
    assert a.data() == b.data()

    # Flusing clears only the flushed backed.
    a.flush()
    assert not a.dirty
    assert b.dirty
    b.flush()
    assert not b.dirty



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server

from . import http
from . import testutil


@pytest.fixture(scope="module")
def daemon():
    daemon = server.Server(config.load(["test/conf/daemon.conf"]))
    daemon.start()
    yield daemon
    daemon.stop()


@pytest.fixture(scope="module")
def proxy():
    proxy = server.Server(config.load(["test/conf/proxy.conf"]))
    proxy.start()
    yield proxy
    proxy.stop()


def test_local_service_disabled(proxy):
    assert proxy.local_service is None


def test_images_extents(daemon, proxy, tmpfile):
    # Get image extents via proxy.
    size = 128 * 1024

    with open(tmpfile, "wb") as f:
        f.truncate(size)

    # Add tickets.
    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size)
    daemon.auth.add(ticket)
    proxy.auth.add(proxy_ticket(daemon, ticket))

    # Get zero extents.
    with http.RemoteClient(proxy.config) as c:
        res = c.request("GET", "/images/{}/extents".format(ticket["uuid"]))
        data = res.read()

    assert res.status == 200

    extents = json.loads(data)
    assert extents == [
        {"start": 0, "length": size, "zero": False, "hole": False}
    ]


@pytest.mark.parametrize("align", [-4096, 0, 4096])
def test_images_download_full(daemon, proxy, tmpfile, align):
    # Simple download of entire image as done by stupid clients.
    size = proxy.config.backend_file.buffer_size + align
    data = b"x" * size

    with open(tmpfile, "wb") as f:
        f.write(data)

    # Add daemon ticket serving tmpfile.
    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size)
    daemon.auth.add(ticket)

    # Add proxy ticket, proxying request to daemon.
    proxy.auth.add(proxy_ticket(daemon, ticket))

    # Download complete image.
    with http.RemoteClient(proxy.config) as c:
        res = c.request("GET", "/images/{}".format(ticket["uuid"]))
        client_data = res.read()

    assert res.status == 200
    assert client_data == data


def test_images_download_options_error(daemon, proxy, tmpfile):
    # Passing OPTIONS error from daemon to proxy client.

    # Create a proxy ticket, but no daemon ticket.
    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=4096)
    proxy.auth.add(proxy_ticket(daemon, ticket))

    # This request should fail in the proxy when opening the backend and
    # sending OPTIONS request.
    with http.RemoteClient(proxy.config) as c:
        res = c.request("GET", "/images/{}".format(ticket["uuid"]))
        res.read()

    # The error should propagate to the caller.
    assert res.status == 403


def test_images_download_get_error(daemon, proxy, tmpfile):
    # Passing GET error from daemon to proxy client.

    size = 128 * 1024
    data = b"x" * size

    with open(tmpfile, "wb") as f:
        f.write(data)

    # Add a daemon ticket with smaller size to trigger 416 error.
    dt = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size - 4096)
    daemon.auth.add(dt)

    # Add proxy ticket with correct size to ensure the request will fail in the
    # daemon.
    pt = proxy_ticket(daemon, dt)
    pt["size"] = size
    proxy.auth.add(pt)

    # This request should fail in the daemon when trying to read after the
    # ticket size.
    with http.RemoteClient(proxy.config) as c:
        res = c.request("GET", "/images/{}".format(dt["uuid"]))
        res.read()

    # The error should propagate to the caller.
    assert res.status == 416


def test_images_download_missing_image(daemon, proxy):
    # Passing GET /extents error from daemon to proxy client.

    # Add a daemon ticket with non-existing file, causing a failure when
    # opening the backend in the daemon.
    ticket = testutil.create_ticket(
        url="file:///no/such/image",
        size=128 * 1024)

    daemon.auth.add(ticket)
    proxy.auth.add(proxy_ticket(daemon, ticket))

    with http.RemoteClient(proxy.config) as c:
        res = c.request("GET", "/images/{}".format(ticket["uuid"]))
        res.read()

    # Pass the daemon error to proxy client.
    assert res.status == 500


@pytest.mark.parametrize("align", [-4096, 0, 4096])
def test_images_upload_full(daemon, proxy, tmpfile, align):
    # Simple upload of entire image as done by stupid clients.
    size = proxy.config.backend_file.buffer_size + align
    data = b"x" * size

    # Create empty sparse image.
    with open(tmpfile, "wb") as f:
        f.truncate(size)

    # Add daemon ticket serving tmpfile.
    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size)
    daemon.auth.add(ticket)

    # Add proxy ticket, proxying request to daemon.
    proxy.auth.add(proxy_ticket(daemon, ticket))

    # Upload data to image.
    with http.RemoteClient(proxy.config) as c:
        res = c.request(
            "PUT",
            "/images/{}".format(ticket["uuid"]),
            body=data)
        res.read()

    assert res.status == 200

    with open(tmpfile, "rb") as f:
        assert f.read() == data


def test_images_upload_error(daemon, proxy, tmpfile):
    # Pass PUT error from daemon to proxy client.

    size = 128 * 1024
    data = b"x" * size

    # Create empty sparse image.
    with open(tmpfile, "wb") as f:
        f.truncate(size)

    # Add a daemon ticket with smaller size to trigger 416 error.
    dt = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size - 4096)
    daemon.auth.add(dt)

    # Add proxy ticket with correct size to ensure the request will fail in the
    # daemon.
    pt = proxy_ticket(daemon, dt)
    pt["size"] = size
    proxy.auth.add(pt)

    # Upload data to image.
    with http.RemoteClient(proxy.config) as c:
        res = c.request(
            "PUT",
            "/images/{}".format(dt["uuid"]),
            body=data)
        res.read()

    assert res.status == 416


def test_images_zero(daemon, proxy, tmpfile):
    # Zero entire image via proxy.
    size = 128 * 1024

    # Create image with data.
    with open(tmpfile, "wb") as f:
        f.write(b"x" * size)

    # Add tickets.
    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size)
    daemon.auth.add(ticket)
    proxy.auth.add(proxy_ticket(daemon, ticket))

    # Zero entire image.
    patch = {"op": "zero", "offset": 0, "size": size}
    body = json.dumps(patch).encode("utf-8")

    with http.RemoteClient(proxy.config) as c:
        res = c.request(
            "PATCH",
            "/images/{}".format(ticket["uuid"]),
            body=body)
        res.read()

    assert res.status == 200

    with open(tmpfile, "rb") as f:
        assert f.read() == b"\0" * size


def test_images_zero_error(daemon, proxy):
    # Pass PATCH error from daemon to proxy client.
    size = 128 * 1024

    # Add tickets for non existing image.
    ticket = testutil.create_ticket(
        url="file:///no/such/image",
        size=size)
    daemon.auth.add(ticket)
    proxy.auth.add(proxy_ticket(daemon, ticket))

    # Zero entire image.
    patch = {"op": "zero", "offset": 0, "size": size}
    body = json.dumps(patch).encode("utf-8")

    with http.RemoteClient(proxy.config) as c:
        res = c.request(
            "PATCH",
            "/images/{}".format(ticket["uuid"]),
            body=body)
        res.read()

    assert res.status == 500


def test_images_flush(daemon, proxy, tmpfile):
    # Zero entire image via proxy.
    size = 128 * 1024

    # Create empty sparse image.
    with open(tmpfile, "wb") as f:
        f.truncate(size)

    # Add tickets.
    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size)
    daemon.auth.add(ticket)
    proxy.auth.add(proxy_ticket(daemon, ticket))

    # Flush image.
    patch = {"op": "flush"}
    body = json.dumps(patch).encode("utf-8")

    with http.RemoteClient(proxy.config) as c:
        res = c.request(
            "PATCH",
            "/images/{}".format(ticket["uuid"]),
            body=body)
        res.read()

    assert res.status == 200


@pytest.mark.parametrize("conf_files", [
    # Using system PKI for both proxy and daemon (default oVirt setup).
    pytest.param(["test/conf/proxy.conf"], id="system"),
    # Using user PKI for proxy and system PKI for daemon (oVirt setup with user
    # certificates).
    pytest.param(
        ["test/conf/proxy.conf", "test/conf/user-tls.conf"], id="user")
])
def test_tls(daemon, tmpfile, conf_files):
    size = daemon.config.backend_file.buffer_size
    data = b"x" * size

    with open(tmpfile, "wb") as f:
        f.write(data)

    # Add daemon ticket serving tmpfile.
    ticket = testutil.create_ticket(
        url="file://{}".format(tmpfile),
        size=size)
    daemon.auth.add(ticket)

    proxy = server.Server(config.load(conf_files))
    proxy.start()
    try:
        # Add proxy ticket, proxying request to daemon.
        proxy.auth.add(proxy_ticket(daemon, ticket))

        # Download complete image.
        with http.RemoteClient(proxy.config) as c:
            res = c.request("GET", "/images/{}".format(ticket["uuid"]))
            client_data = res.read()
    finally:
        proxy.stop()

    assert res.status == 200
    assert client_data == data


def proxy_ticket(daemon, daemon_ticket):
    """
    Create a proxy ticket from daemon ticket.
    """
    ticket = dict(daemon_ticket)

    host, port = daemon.remote_service.address
    ticket["url"] = "https://{}:{}/images/{}".format(
        host, port, ticket["uuid"])

    return ticket



# ovirt-imageio
# Copyright (C) 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from ovirt_imageio._internal.measure import Range, RangeList

# Compare ranges.


def test_range_lt_start_smaller():
    assert Range(5, 10) < Range(6, 10)
    assert not Range(6, 10) < Range(5, 10)


def test_range_lt_start_same():
    assert Range(5, 10) < Range(5, 11)
    assert not Range(5, 11) < Range(5, 10)


def test_range_eq():
    assert Range(5, 10) == Range(5, 10)


def test_ragne_ne():
    assert Range(5, 10) != Range(15, 20)


# Add to range list


def test_range_list_empty():
    rl = RangeList()
    assert rl.sum() == 0


def test_range_list_add_first():
    rl = RangeList()
    rl.add(Range(0, 10))
    assert rl._ranges == [Range(0, 10)]
    assert rl.sum() == 10


def test_range_list_add_same():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(0, 10))
    assert rl._ranges == [Range(0, 10)]
    assert rl.sum() == 10


def test_range_list_add_shorter():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(0, 5))
    assert rl._ranges == [Range(0, 10)]
    assert rl.sum() == 10


def test_range_list_add_longer():
    rl = RangeList()
    rl.add(Range(0, 5))
    rl.add(Range(0, 10))
    assert rl._ranges == [Range(0, 10)]
    assert rl.sum() == 10


def test_range_list_add_overlap():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(5, 15))
    assert rl._ranges == [Range(0, 15)]
    assert rl.sum() == 15


def test_range_list_add_contiguous():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(10, 20))
    assert rl._ranges == [Range(0, 20)]
    assert rl.sum() == 20


def test_range_list_add_non_contiguous():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(20, 30))
    assert rl._ranges == [Range(0, 10), Range(20, 30)]
    assert rl.sum() == 20


def test_ragne_list_add_overlap_next_some():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(20, 30))
    rl.add(Range(15, 25))
    assert rl._ranges == [Range(0, 10), Range(15, 30)]
    assert rl.sum() == 25


def test_ragne_list_add_overlap_next_all():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(20, 30))
    rl.add(Range(15, 30))
    assert rl._ranges == [Range(0, 10), Range(15, 30)]
    assert rl.sum() == 25


def test_range_list_add_overlap_both_some():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(20, 30))
    rl.add(Range(5, 25))
    assert rl._ranges == [Range(0, 30)]
    assert rl.sum() == 30


def test_range_list_add_overlap_both_all():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(20, 30))
    rl.add(Range(0, 30))
    assert rl._ranges == [Range(0, 30)]
    assert rl.sum() == 30


def test_range_list_add_overlap_next_multi_some():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(20, 30))
    rl.add(Range(40, 50))
    rl.add(Range(60, 70))
    rl.add(Range(25, 45))
    assert rl._ranges == [Range(0, 10), Range(20, 50), Range(60, 70)]
    assert rl.sum() == 50


def test_range_list_add_overlap_next_multi_all():
    rl = RangeList()
    rl.add(Range(0, 10))
    rl.add(Range(20, 30))
    rl.add(Range(40, 50))
    rl.add(Range(60, 70))
    rl.add(Range(20, 50))
    assert rl._ranges == [Range(0, 10), Range(20, 50), Range(60, 70)]
    assert rl.sum() == 50


# Update range list.


def test_range_list_update_first():
    rl = RangeList()
    rl.update([Range(0, 10), Range(10, 20), Range(20, 30), Range(30, 40)])
    assert rl._ranges == [Range(0, 40)]
    assert rl.sum() == 40


def test_range_list_update_same():
    rl = RangeList()
    rl.add(Range(0, 40))
    rl.update([Range(0, 10), Range(10, 20), Range(20, 30), Range(30, 40)])
    assert rl._ranges == [Range(0, 40)]
    assert rl.sum() == 40


def test_range_list_update_contiguous():
    rl = RangeList()
    rl.update([Range(0, 10), Range(100, 110), Range(200, 210)])
    rl.update([Range(10, 20), Range(110, 120), Range(210, 220)])
    assert rl._ranges == [Range(0, 20), Range(100, 120), Range(200, 220)]
    assert rl.sum() == 60


def test_range_list_update_non_contiguous():
    rl = RangeList()
    rl.update([Range(0, 10), Range(100, 110), Range(200, 210)])
    rl.update([Range(300, 310), Range(400, 410), Range(500, 510)])
    assert rl._ranges == [
        Range(0, 10),
        Range(100, 110),
        Range(200, 210),
        Range(300, 310),
        Range(400, 410),
        Range(500, 510),
    ]
    assert rl.sum() == 60


def test_ragne_list_update_overlap_next_some():
    rl = RangeList()
    rl.update([Range(0, 10), Range(20, 30)])
    rl.update([Range(15, 25)])
    assert rl._ranges == [Range(0, 10), Range(15, 30)]
    assert rl.sum() == 25


def test_ragne_list_update_overlap_next_all():
    rl = RangeList()
    rl.update([Range(0, 10), Range(20, 30)])
    rl.update([Range(15, 30)])
    assert rl._ranges == [Range(0, 10), Range(15, 30)]
    assert rl.sum() == 25


def test_range_list_update_overlap_both_some():
    rl = RangeList()
    rl.update([Range(0, 10), Range(20, 30)])
    rl.update([Range(5, 25)])
    assert rl._ranges == [Range(0, 30)]
    assert rl.sum() == 30


def test_range_list_update_overlap_both_all():
    rl = RangeList()
    rl.update([Range(0, 10), Range(20, 30)])
    rl.update([Range(0, 30)])
    assert rl._ranges == [Range(0, 30)]
    assert rl.sum() == 30


def test_range_list_update_overlap_next_multi_some():
    rl = RangeList()
    rl.update([Range(0, 10), Range(20, 30), Range(40, 50), Range(60, 70)])
    rl.update([Range(5, 35), Range(25, 65)])
    assert rl._ranges == [Range(0, 70)]
    assert rl.sum() == 70


def test_range_list_update_overlap_next_multi_all():
    rl = RangeList()
    rl.update([Range(0, 10), Range(20, 30), Range(40, 50), Range(60, 70)])
    rl.update([Range(0, 50), Range(20, 70)])
    assert rl._ranges == [Range(0, 70)]
    assert rl.sum() == 70


# Copy range list.


def test_range_list_copy():
    r1 = RangeList()

    r1.add(Range(0, 100))
    r1.add(Range(200, 300))
    r1.add(Range(400, 500))

    r2 = RangeList(r1)

    # r1 and r2 are qqual.
    assert r1._ranges == r2._ranges
    assert r1.sum() == r2.sum()

    # But independent.
    r1.add(Range(300, 400))
    r2.add(Range(100, 200))

    assert r1._ranges == [Range(0, 100), Range(200, 500)]
    assert r2._ranges == [Range(0, 300), Range(400, 500)]



# ovirt-imageio
# Copyright (C) 2015-2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import argparse
import configparser
import glob
import grp
import json
import logging
import logging.config
import os
import pwd
import signal
import socket
import sys

import systemd.daemon

from . import auth
from . import config
from . import services
from . import version

DEFAULT_CONF_DIR = "/etc/ovirt-imageio"
VENDOR_CONF_DIR = "/usr/lib/ovirt-imageio"

log = logging.getLogger("server")


def main():
    args = parse_args()
    try:
        cfg = load_config(args.conf_dir)
        if args.show_config:
            show_config(cfg)
            return

        configure_logger(cfg)
        log.info("Starting (hostname=%s pid=%s, version=%s)",
                 socket.gethostname(), os.getpid(), version.string)

        server = Server(cfg)
        signal.signal(signal.SIGINT, server.terminate)
        signal.signal(signal.SIGTERM, server.terminate)

        server.start()
        try:
            systemd.daemon.notify("READY=1")
            log.info("Ready for requests")
            while server.running:
                signal.pause()
        finally:
            server.stop()
        log.info("Stopped")
    except Exception:
        log.exception("Server failed")
        sys.exit(1)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c", "--conf-dir",
        default=DEFAULT_CONF_DIR,
        help="path to configuration directory, where daemon.conf and "
             "logger.conf are located.")
    parser.add_argument(
        "--show-config",
        action="store_true",
        help="print actual configuration in json format and exit. This is "
             "useful for debugging configuration issues, or reading imageio "
             "configuration by other programs.")
    return parser.parse_args()


def show_config(cfg):
    print(json.dumps(config.to_dict(cfg), indent=4))


def load_config(conf_dir):
    pattern = os.path.join(conf_dir, "conf.d", "*.conf")
    files = glob.glob(pattern)

    # Trying to use invalid configuration directory is a user error or
    # broken installation. Failing fast will help to fix the issue.
    # https://github.com/oVirt/ovirt-imageio/issues/33
    if not files:
        raise ValueError(f"Could not find {pattern}")

    # Vendor may override application defaults if needed.
    pattern = os.path.join(VENDOR_CONF_DIR, "conf.d", "*.conf")
    files.extend(glob.glob(pattern))

    # Override files based on file name sort order:
    # - /var/lib/ovirt-imageio/conf.d/75-vendor.conf overrides
    #   /etc/ovirt-imageio/conf.d/50-vdsm.conf.
    # - /etc/ovirt-imageio/conf.d/99-user.conf overrides
    #   /var/lib/ovirt-imageio/conf.d/75-vendor.conf.
    files.sort(key=os.path.basename)

    return config.load(files)


def configure_logger(cfg):
    parser = configparser.RawConfigParser()
    parser.read_dict(config.to_dict(cfg))
    logging.config.fileConfig(parser, disable_existing_loggers=False)


class Server:

    def __init__(self, config):
        self.config = config
        self.running = False
        self.auth = auth.Authorizer(config)
        self.remote_service = services.RemoteService(self.config, self.auth)
        self.local_service = None
        if config.local.enable:
            self.local_service = services.LocalService(self.config, self.auth)
        self.control_service = services.ControlService(self.config, self.auth)

        if os.geteuid() == 0 and self.config.daemon.drop_privileges:
            self._drop_privileges()

    def start(self):
        assert not self.running
        self.running = True

        self.remote_service.start()
        if self.local_service is not None:
            self.local_service.start()
        self.control_service.start()

    def stop(self):
        log.debug("Stopping services")
        self.remote_service.stop()
        if self.local_service is not None:
            self.local_service.stop()
        self.control_service.stop()

    def terminate(self, signo, frame):
        log.info("Received signal %d, shutting down", signo)
        self.running = False

    def _drop_privileges(self):
        uid = pwd.getpwnam(self.config.daemon.user_name).pw_uid
        gid = grp.getgrnam(self.config.daemon.group_name).gr_gid

        # Restore ownership of run directory.
        run_dir = self.config.daemon.run_dir
        if os.path.exists(run_dir):
            log.debug(
                "Changing ownership of %s to %i:%i" % (run_dir, uid, gid))
            os.chown(run_dir, uid, gid)

        # Restore ownership of log files.
        for h in logging.root.handlers:
            # Support only logging.FileHandler and sub-classes.
            filename = getattr(h, "baseFilename", None)
            if filename is not None:
                log.debug(
                    "Changing ownership of %s to %i:%i" % (filename, uid, gid))
                os.chown(filename, uid, gid)

        # Restore ownership of control socket is used.
        transport = self.config.control.transport.lower()
        if transport == "unix":
            os.chown(self.config.control.socket, uid, gid)

        # Set new uid and gid for the process.
        log.debug("Dropping root privileges, running as %i:%i" % (uid, gid))
        os.initgroups(self.config.daemon.user_name, gid)
        os.setgid(gid)
        os.setuid(uid)



# ovirt-imageio
# Copyright (C) 2021 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
imageio admin library.
"""

import http.client
import json

from .. _internal import uhttp


class Error(Exception):
    """
    Base class for admin library errors.
    """


class ClientError(Error):
    """
    Client failed to communicate with the server.
    """


class ServerError(Error):
    """
    Server request failed.
    """
    def __init__(self, code, message):
        # The HTTP status returned by the server.
        self.code = code

        # The HTTP error message retrned by the server.
        self.message = message

    def __str__(self):
        return f"Server error: {self.code} {self.message}"


class Client:
    """
    imageio admin client.
    """

    def __init__(self, cfg, timeout=60):
        """
        Create a client using specificed configuration.

        Arguments:
          cfg (dict): Server configuration, loaded using admin.load_config().
          timeout (float): Connection timeout in seconds (default 60).
        """
        transport = cfg.control.transport.lower()
        if transport == "tcp":
            self.con = http.client.HTTPConnection(
                "localhost", cfg.control.port, timeout=timeout)
        elif transport == "unix":
            self.con = uhttp.UnixHTTPConnection(
                cfg.control.socket, timeout=timeout)
        else:
            raise ValueError(f"Invalid control.transport: {transport}")

    def add_ticket(self, ticket):
        """
        Add a ticket to imageio daemon.

        Arguments:
          ticket (dict): Ticket to add to running imageio daemon.
        """
        status, body = self._request(
            "PUT",
            f"/tickets/{ticket['uuid']}",
            body=json.dumps(ticket).encode("utf-8"))
        if status != http.client.OK:
            raise ServerError(status, body)

    def get_ticket(self, ticket_id):
        """
        Get ticket information from imageio daemon.

        Arguments:
          ticket_id (str): Ticket id.
        """
        status, body = self._request("GET", f"/tickets/{ticket_id}")
        if status != http.client.OK:
            raise ServerError(status, body)
        return json.loads(body)

    def mod_ticket(self, ticket_id, changes):
        """
        Modify a ticket in imageio daemon.

        Currenlty the only property that may be changed is "timeout".

        Arguments:
          ticket_id (str): Ticket id.
          changes (dict): Ticket properties to modify.
        """
        status, body = self._request(
            "PATCH",
            f"/tickets/{ticket_id}",
            body=json.dumps(changes).encode("utf-8"))
        if status != http.client.OK:
            raise ServerError(status, body)

    def del_ticket(self, ticket_id):
        """
        Delete a ticket from imageio daemon.

        Arguments:
          ticket_id (str): Ticket id.
        """
        status, body = self._request("DELETE", f"/tickets/{ticket_id}")
        if status != http.client.NO_CONTENT:
            raise ServerError(status, body)

    def start_profile(self):
        status, body = self._request("POST", "/profile/?run=y")
        if status != http.client.OK:
            raise ServerError(status, body)

    def stop_profile(self):
        status, body = self._request("POST", "/profile/?run=n")
        if status != http.client.OK:
            raise ServerError(status, body)

    def close(self):
        """
        Close the conection to imageio daemon.
        """
        self.con.close()

    def _request(self, method, uri, body=None):
        try:
            self.con.request(method, uri, body=body)
            res = self.con.getresponse()
            body = res.read().decode("utf-8")
            return res.status, body
        except Exception as e:
            raise ClientError(str(e))

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pytest
from ovirt_imageio._internal import stats


class FakeTime:

    def __init__(self):
        self.value = 0

    def __call__(self):
        return self.value


# Ccorrect usage

def test_empty():
    c = stats.Clock()
    assert str(c) == ""


def test_stop_returns_elapsed_time():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)

    c.start("read")
    fake_time.value += 1
    assert c.stop("read") == 1

    c.start("read")
    fake_time.value += 2
    assert c.stop("read") == 2


def test_measure():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)
    c.start("total")
    c.start("read")
    fake_time.value += 1
    c.stop("read")
    c.start("write")
    fake_time.value += 1
    c.stop("write")
    c.start("sync")
    fake_time.value += 1
    c.stop("sync")
    c.stop("total")
    assert str(c) == (
        "[total 1 ops, 3.000000 s] "
        "[read 1 ops, 1.000000 s] "
        "[write 1 ops, 1.000000 s] "
        "[sync 1 ops, 1.000000 s]"
    )


def test_measure_multiple():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)
    c.start("total")
    c.start("read")
    fake_time.value += 1
    c.stop("read")
    c.start("write")
    fake_time.value += 1
    c.stop("write")
    c.start("read")
    fake_time.value += 1
    c.stop("read")
    c.start("write")
    fake_time.value += 1
    c.stop("write")
    c.start("sync")
    fake_time.value += 1
    c.stop("sync")
    c.stop("total")
    assert str(c) == (
        "[total 1 ops, 5.000000 s] "
        "[read 2 ops, 2.000000 s] "
        "[write 2 ops, 2.000000 s] "
        "[sync 1 ops, 1.000000 s]"
    )


def test_running():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)
    c.start("total")
    fake_time.value += 3
    c.start("read")
    fake_time.value += 4
    c.stop("read")
    # Since total was not stopped, the operation is not counted.
    assert str(c) == "[total 0 ops, 7.000000 s] [read 1 ops, 4.000000 s]"


def test_bytes():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)

    c.start("total")

    # Using start()/stop().

    s = c.start("read")
    s.bytes += 512 * 1024**2
    fake_time.value += 1
    c.stop("read")

    s = c.start("write")
    s.bytes += 200 * 1024**2
    fake_time.value += 1
    c.stop("write")

    # Using run().

    with c.run("read") as s:
        s.bytes += 512 * 1024**2
        fake_time.value += 1

    with c.run("write") as s:
        s.bytes += 200 * 1024**2
        fake_time.value += 1

    with c.run("sync"):
        fake_time.value += 1

    c.stop("total")

    assert str(c) == (
        "[total 1 ops, 5.000000 s] "
        "[read 2 ops, 2.000000 s, 1.00 GiB, 512.00 MiB/s] "
        "[write 2 ops, 2.000000 s, 400.00 MiB, 200.00 MiB/s] "
        "[sync 1 ops, 1.000000 s]"
    )


def test_null_clock():
    c = stats.NullClock()

    # This clock records nothing...
    with c.run("total"):
        with c.run("read") as s:
            s.bytes += 4096
            # bytes value is dropped...
            assert s.bytes == 0
        with c.run("sync"):
            pass

    # And always return empty string.
    assert str(c) == ""


def test_error_before_stop():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)
    c.start("read")
    fake_time.value += 1

    # User code fails here, before we increase bytes and call stop().

    # Since read was not stopped the operation is not counted...
    assert str(c) == "[read 0 ops, 1.000000 s]"

    # And we cannot start this operation again.
    with pytest.raises(RuntimeError):
        c.start("read")

    # We can abort the operation to continue to use this clock.
    fake_time.value += 1
    c.abort("read")
    assert str(c) == "[read 0 ops, 2.000000 s]"


def test_error_in_run():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)

    with pytest.raises(RuntimeError):
        with c.run("read") as s:
            fake_time.value += 1
            raise RuntimeError("fake error")

            s.bytes += 512 * 1024**2

    # First read failed, so bytes is not set and the operation is not counted.
    assert str(c) == "[read 0 ops, 1.000000 s]"

    # However clock was stopped, so next read can succeed.
    with c.run("read") as s:
        fake_time.value += 1
        s.bytes += 512 * 1024**2

    # Bytes value is set now so we report total and rate values.
    assert str(c) == "[read 1 ops, 2.000000 s, 512.00 MiB, 256.00 MiB/s]"


# Inccorrect usage

def test_start_twice():
    c = stats.Clock()
    c.start("started")
    with pytest.raises(RuntimeError):
        c.start("started")


def test_stop_twice():
    c = stats.Clock()
    c.start("stopped")
    c.stop("stopped")
    with pytest.raises(RuntimeError):
        c.stop("stopped")


def test_stop_missing():
    c = stats.Clock()
    with pytest.raises(RuntimeError):
        c.stop("missing")


def test_run():
    fake_time = FakeTime()
    c = stats.Clock(fake_time)
    with c.run("total"):
        with c.run("a"):
            fake_time.value += 4
        with c.run("b"):
            fake_time.value += 3
    assert str(c) == (
        "[total 1 ops, 7.000000 s] "
        "[a 1 ops, 4.000000 s] "
        "[b 1 ops, 3.000000 s]"
    )


def test_run_recursive():
    c = stats.Clock()
    with c.run("started"):
        with pytest.raises(RuntimeError):
            with c.run("started"):
                pass


@pytest.mark.benchmark
@pytest.mark.parametrize("clock", [
    pytest.param(stats.Clock(), id="clock"),
    pytest.param(stats.NullClock(), id="null-clock"),
])
def test_benchmark(clock):
    test = stats.Clock()
    with test.run("total"):
        clock.start("connection")
        # We have seen 66,000 requests per single upload with virt-v2v.
        for i in range(50000):
            clock.start("request")
            with clock.run("read") as s:
                s.bytes += 2 * 1024**2
            with clock.run("write") as s:
                s.bytes += 2 * 1024**2
            clock.stop("request")
        clock.stop("connection")
    print("{} {}".format(test, clock))



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


def is_fedora(version=""):
    return _check("Fedora release {}".format(version))


def is_centos(version=""):
    return _check("CentOS Stream release {}".format(version))


def is_rhel(version=""):
    return _check("Red Hat Enterprise Linux release {}".format(version))


def _check(text):
    with open("/etc/redhat-release") as f:
        line = f.readline()
    return text in line



# ovirt-imageio
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import logging
import threading
import urllib.parse as urllib_parse

from . import backends
from . import errors
from . import measure
from . import ops
from . import util

log = logging.getLogger("auth")


class Ticket:

    def __init__(self, ticket_dict, cfg):
        if not isinstance(ticket_dict, dict):
            raise errors.InvalidTicket(
                "Invalid ticket: %r, expecting a dict" % ticket_dict)

        self._uuid = _required(ticket_dict, "uuid", str)
        self._size = _required(ticket_dict, "size", int)
        self._ops = _required(ticket_dict, "ops", list)

        self._timeout = _required(ticket_dict, "timeout", int)
        self._inactivity_timeout = _optional(
            ticket_dict, "inactivity_timeout", int,
            default=cfg.daemon.inactivity_timeout)

        now = int(util.monotonic_time())
        self._expires = now + self._timeout
        self._access_time = now

        url_str = _required(ticket_dict, "url", str)
        try:
            self._url = urllib_parse.urlparse(url_str)
        except (ValueError, AttributeError, TypeError) as e:
            raise errors.InvalidTicketParameter("url", url_str, e)
        if not backends.supports(self._url.scheme):
            raise errors.InvalidTicketParameter(
                "url", url_str,
                "Unsupported url scheme: %s" % self._url.scheme)

        self._transfer_id = _optional(ticket_dict, "transfer_id", str)

        # Engine before 4.2.7 did not pass the transfer id. Generate a likey
        # unique value from the first half of the uuid.
        if self._transfer_id is None:
            self._transfer_id = f"(ticket/{self._uuid[:18]})"

        self._filename = _optional(ticket_dict, "filename", str)
        self._sparse = _optional(ticket_dict, "sparse", bool, default=False)
        self._dirty = _optional(ticket_dict, "dirty", bool, default=False)

        self._operations = []
        self._lock = threading.Lock()

        # Set holding ongoing operations.
        self._ongoing = set()

        # Ranges transferred by completed operations.
        self._completed = measure.RangeList()

        # Set to true when a ticket is canceled. Once canceled, all operations
        # on this ticket will raise errors.AuthorizationError.
        self._canceled = False

        # Mapping of connection id to connection context. When empty, this
        # ticket is not used by any connection.
        self._connections = {}

        # Used for waiting until a ticket is unused during cancellation. A
        # ticket can be removed only when this event is set.
        self._unused = threading.Event()

    @property
    def uuid(self):
        return self._uuid

    @property
    def size(self):
        return self._size

    @property
    def url(self):
        return self._url

    @property
    def ops(self):
        return self._ops

    @property
    def expires(self):
        return self._expires

    @property
    def transfer_id(self):
        """
        Return the ticket transfer id, available since engine 4.2.7 or None
        if the ticket was generated by older engine.
        """
        return self._transfer_id

    @property
    def filename(self):
        return self._filename

    @property
    def sparse(self):
        return self._sparse

    @property
    def dirty(self):
        """
        Return True if ticket's url should provide dirty extents information.
        """
        return self._dirty

    @property
    def idle_time(self):
        """
        Return the time in which the ticket became inactive.
        """
        if self.active():
            return 0
        return int(util.monotonic_time()) - self._access_time

    @property
    def inactivity_timeout(self):
        """
        Return the number of seconds to wait before disconnecting inactive
        client.
        """
        return self._inactivity_timeout

    @property
    def canceled(self):
        with self._lock:
            return self._canceled

    def add_context(self, con_id, context):
        with self._lock:
            if self._canceled:
                raise errors.AuthorizationError(
                    "Transfer {} was canceled".format(self.transfer_id))

            log.debug("Adding connection %s context to transfer %s",
                      con_id, self.transfer_id)
            self._connections[con_id] = context

    def get_context(self, con_id):
        return self._connections[con_id]

    def remove_context(self, con_id):
        with self._lock:
            try:
                context = self._connections[con_id]
            except KeyError:
                return

            log.debug("Removing connection %s context from transfer %s",
                      con_id, self.transfer_id)
            context.close()

            # If context was closed, it is safe to remove it.
            del self._connections[con_id]

    def run(self, operation):
        """
        Run an operation, binding it to the ticket.
        """
        self._add_operation(operation)
        try:
            return operation.run()
        except ops.Canceled:
            log.debug("Operation %s was canceled", operation)
        finally:
            self._remove_operation(operation)

    def touch(self):
        """
        Extend the ticket and update the last access time.

        Must be called when an operation is completed.
        """
        now = int(util.monotonic_time())
        self._expires = now + self._timeout
        self._access_time = now

    def _add_operation(self, op):
        with self._lock:
            if self._canceled:
                raise errors.AuthorizationError(
                    "Transfer {} was canceled".format(self.transfer_id))

            self._ongoing.add(op)

    def _remove_operation(self, op):
        with self._lock:
            self._ongoing.remove(op)

            if self._canceled:
                # If this was the last ongoing operation, wake up caller
                # waiting on cancel().
                if not self._ongoing:
                    log.debug(
                        "Removed last ongoring operation for transfer %s",
                        self.transfer_id)
                    self._unused.set()

                raise errors.AuthorizationError(
                    "Transfer {} was canceled".format(self.transfer_id))

            # We don't report transfered bytes for read-write ticket.
            if len(self.ops) == 1:
                r = measure.Range(op.offset, op.offset + op.done)
                self._completed.add(r)

        self.touch()

    def active(self):
        return bool(self._ongoing)

    def transferred(self):
        """
        The number of bytes that were transferred so far using this ticket.
        """
        if len(self.ops) > 1:
            # Both read and write, cannot report meaningful value.
            return None

        with self._lock:
            # NOTE: this must not modify the ticket state.
            completed = measure.RangeList(self._completed)
            ongoing = [measure.Range(op.offset, op.offset + op.done)
                       for op in self._ongoing]

        completed.update(ongoing)
        return completed.sum()

    def may(self, op):
        if op == "read":
            # Having "write" imply also "read".
            return "read" in self.ops or "write" in self.ops
        else:
            return op in self.ops

    def info(self):
        info = {
            "active": self.active(),
            "canceled": self._canceled,
            "connections": len(self._connections),
            "expires": self._expires,
            "idle_time": self.idle_time,
            "inactivity_timeout": self._inactivity_timeout,
            "ops": list(self._ops),
            "size": self._size,
            "sparse": self._sparse,
            "dirty": self._dirty,
            "timeout": self._timeout,
            "url": urllib_parse.urlunparse(self._url),
            "uuid": self._uuid,
        }
        if self._transfer_id:
            info["transfer_id"] = self._transfer_id
        if self.filename:
            info["filename"] = self.filename
        transferred = self.transferred()
        if transferred is not None:
            info["transferred"] = transferred
        return info

    def extend(self, timeout):
        expires = int(util.monotonic_time()) + timeout
        self._expires = expires

    def cancel(self, timeout=60):
        """
        Cancel a ticket and wait until all ongoing operations finish.

        Arguments:
            timeout (float): number of seconds to wait until the ticket is
                unused. If timeout is zero, return immediately without waiting.
                The caller will have to poll the ticket status until the number
                of connections becomes zero.

        Returns:
            True if ticket can be removed.

        Raises:
            errors.TicketCancelTimeout if timeout is non-zero and the ticket is
            still used when the timeout expires.
        """
        log.debug("Cancelling transfer %s", self.transfer_id)

        with self._lock:
            # No operation can start now, and new connections cannot be added
            # to the ticket.
            self._canceled = True

            if not self._ongoing:
                # There are no ongoing opearations, but we may have idle
                # connections - release their resources.
                for ctx in self._connections.values():
                    ctx.close()
                log.debug("Transfer %s was canceled", self.transfer_id)
                return True

            log.debug("Canceling transfer %s ongoing operations",
                      self.transfer_id)
            # Cancel ongoing operations. This speeds up cancellation when
            # streaming lot of data. Operations will be canceled once they
            # complete the current I/O.
            for op in self._ongoing:
                op.cancel()

        if timeout:
            log.info("Waiting until transfer %s ongoing operations finish",
                     self.transfer_id)
            if not self._unused.wait(timeout):
                raise errors.TransferCancelTimeout(self.transfer_id)

            # Finished ongoing operations discover that the ticket was canceled
            # and close the connection. We need to release resources used by
            # idle connections.
            with self._lock:
                for ctx in self._connections.values():
                    ctx.close()

            log.info("Transfer %s was canceled", self.transfer_id)
            return True

        # The caller need to wait until ongoing operations finish by polling
        # the ticket "active" property. When the ticket becomes inactive,
        # caller must call again to delete the ticket.
        return False

    def __repr__(self):
        return ("<Ticket "
                "active={active!r} "
                "canceled={self._canceled} "
                "connections={connections} "
                "expires={self.expires!r} "
                "inactivity_timeout={self.inactivity_timeout} "
                "filename={self.filename!r} "
                "idle_time={self.idle_time} "
                "ops={self.ops!r} "
                "size={self.size!r} "
                "sparse={self.sparse!r} "
                "dirty={self.dirty!r} "
                "transfer_id={self.transfer_id!r} "
                "transferred={transferred!r} "
                "url={url!r} "
                "uuid={self.uuid!r} "
                "at {addr:#x}>"
                ).format(
                    active=self.active(),
                    addr=id(self),
                    connections=len(self._connections),
                    self=self,
                    transferred=self.transferred(),
                    url=urllib_parse.urlunparse(self.url)
                )


def _required(d, key, type):
    if key not in d:
        raise errors.MissingTicketParameter(key)
    return _validate(key, d[key], type)


def _optional(d, key, type, default=None):
    if key not in d:
        return default
    return _validate(key, d[key], type)


def _validate(key, value, type):
    if not isinstance(value, type):
        raise errors.InvalidTicketParameter(
            key, value, "expecting a {!r} value".format(type))
    return value


class Authorizer:

    def __init__(self, config):
        self._config = config
        self._tickets = {}

    def add(self, ticket_dict):
        """
        Add a ticket to the store.

        Raises errors.InvalidTicket if ticket dict is invalid.
        """
        ticket = Ticket(ticket_dict, self._config)
        self._tickets[ticket.uuid] = ticket

    def remove(self, ticket_id):
        try:
            ticket = self._tickets[ticket_id]
        except KeyError:
            log.debug("Ticket %s does not exist", ticket_id)
            return

        # Cancel the ticket and wait until the ticket is unused. Will raise
        # errors.TransferCancelTimeout if the ticket could not be canceled
        # within the timeout.
        if ticket.cancel(self._config.control.remove_timeout):
            # Ticket is unused now, so it is safe to remove it.
            del self._tickets[ticket_id]

    def clear(self):
        self._tickets.clear()

    def get(self, ticket_id):
        """
        Gets a ticket ID and returns the proper
        Ticket object from the tickets' cache.
        """
        return self._tickets[ticket_id]

    def authorize(self, ticket_id, op):
        """
        Authorizing a ticket operation
        """
        try:
            ticket = self._tickets[ticket_id]
        except KeyError:
            raise errors.AuthorizationError(
                "No such ticket {}".format(ticket_id))

        log.debug("AUTH op=%s transfer=%s", op, ticket.transfer_id)

        if ticket.canceled:
            raise errors.AuthorizationError(
                "Transfer={} was canceled".format(ticket.transfer_id))

        if ticket.expires <= util.monotonic_time():
            raise errors.AuthorizationError(
                "Transfer={} expired".format(ticket.transfer_id))

        if not ticket.may(op):
            raise errors.AuthorizationError(
                "Transfer={} forbids {}".format(ticket.transfer_id, op))

        return ticket



# ovirt-imageio
# Copyright (C) 2015-2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
import logging

from .. import backends
from .. import cors
from .. import errors
from .. import http
from .. import ops
from .. import validate

log = logging.getLogger("images")

BASE_FEATURES = ("checksum", "extents")
ALL_FEATURES = BASE_FEATURES + ("flush", "zero")


class Handler:
    """
    Handle requests for the /images/ resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    @cors.allow()
    def put(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # Used by engine UI to close connection after upload.
        close = req.query.get("close") == "y"
        if close:
            resp.close_connection()

        size = req.content_length
        if size is None:
            raise http.Error(
                http.BAD_REQUEST, "Content-Length header is required")

        offset = req.content_range.first if req.content_range else 0

        # For backward compatibility, we flush by default.
        flush = validate.enum(req.query, "flush", ("y", "n"), default="y")
        flush = (flush == "y")

        try:
            ticket = self.auth.authorize(ticket_id, "write")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        validate.allowed_range(offset, size, ticket)

        log.debug(
            "[%s] WRITE size=%d offset=%d flush=%s close=%s transfer=%s",
            req.client_addr, size, offset, flush, close, ticket.transfer_id)

        op = ops.Write(
            ctx.backend,
            req,
            ctx.buffer,
            size,
            offset=offset,
            flush=flush,
            clock=req.clock)
        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None
        except errors.PartialContent as e:
            raise http.Error(http.BAD_REQUEST, str(e))

    @cors.allow()
    def get(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # Used by engine UI to close connection after download.
        close = req.query.get("close") == "y"
        if close:
            resp.close_connection()

        offset = 0
        size = None
        if req.range:
            offset = req.range.first
            if offset < 0:
                # TODO: support suffix-byte-range-spec "bytes=-last".
                # See https://tools.ietf.org/html/rfc7233#section-2.1.
                raise http.Error(
                    http.REQUESTED_RANGE_NOT_SATISFIABLE,
                    "suffix-byte-range-spec not supported yet")

            if req.range.last is not None:
                # Add 1 to include the last byte in the payload.
                size = req.range.last - offset + 1
                # TODO: validate size with actual image size.

        try:
            ticket = self.auth.authorize(ticket_id, "read")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        if size is not None:
            validate.allowed_range(offset, size, ticket)
            validate.available_range(offset, size, ticket, ctx.backend)
        else:
            size = min(ticket.size, ctx.backend.size()) - offset

        log.debug(
            "[%s] READ size=%d offset=%d close=%s transfer=%s",
            req.client_addr, size, offset, close, ticket.transfer_id)

        content_disposition = "attachment"
        if ticket.filename:
            content_disposition += "; filename=%s" % ticket.filename

        resp.headers["content-length"] = size
        resp.headers["content-type"] = "application/octet-stream"
        resp.headers["content-disposition"] = content_disposition

        if req.range:
            resp.status_code = http.PARTIAL_CONTENT
            resp.headers["content-range"] = "bytes %d-%d/%d" % (
                offset, offset + size - 1, ticket.size)

        op = ops.Read(
            ctx.backend,
            resp,
            ctx.buffer,
            size,
            offset=offset,
            clock=req.clock)
        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None
        except errors.PartialContent as e:
            raise http.Error(http.BAD_REQUEST, str(e))

    def patch(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # TODO: Reject requests with too big payloads. We know the maximum size
        # of a payload based on the keys and the size of offset and size.
        try:
            msg = json.loads(req.read())
        except ValueError as e:
            raise http.Error(
                http.BAD_REQUEST, "Invalid JSON message {}" .format(e))

        op = validate.enum(msg, "op", ("zero", "flush"))
        if op == "zero":
            return self._zero(req, resp, ticket_id, msg)
        elif op == "flush":
            return self._flush(req, resp, ticket_id, msg)
        else:
            raise RuntimeError("Unreachable")

    def _zero(self, req, resp, ticket_id, msg):
        size = validate.integer(msg, "size", minval=0)
        offset = validate.integer(msg, "offset", minval=0, default=0)
        flush = validate.boolean(msg, "flush", default=False)

        try:
            ticket = self.auth.authorize(ticket_id, "write")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        validate.allowed_range(offset, size, ticket)

        log.debug(
            "[%s] ZERO size=%d offset=%d flush=%s transfer=%s",
            req.client_addr, size, offset, flush, ticket.transfer_id)

        op = ops.Zero(
            ctx.backend,
            size,
            offset=offset,
            flush=flush,
            clock=req.clock)

        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None
        except errors.PartialContent as e:
            raise http.Error(http.BAD_REQUEST, str(e))

    def _flush(self, req, resp, ticket_id, msg):
        try:
            ticket = self.auth.authorize(ticket_id, "write")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        log.info("[%s] FLUSH transfer=%s",
                 req.client_addr, ticket.transfer_id)

        op = ops.Flush(ctx.backend, clock=req.clock)

        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None

    @cors.allow(allow_methods="OPTIONS,GET,PUT")
    def options(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        options = {}

        if self.config.local.enable:
            options["unix_socket"] = self.config.local.socket

        if ticket_id == "*":
            log.info("[%s] OPTIONS transfer=*", req.client_addr)
            # Reporting the meta-capabilities for all images.
            allow = ["OPTIONS", "GET", "PUT", "PATCH"]
            options["features"] = ALL_FEATURES
        else:
            # Reporting real image capabilities per ticket.
            try:
                ticket = self.auth.authorize(ticket_id, "read")
                ctx = backends.get(req, ticket, self.config)
            except errors.AuthorizationError as e:
                resp.close_connection()
                raise http.Error(http.FORBIDDEN, str(e))

            log.info("[%s] OPTIONS transfer=%s",
                     req.client_addr, ticket.transfer_id)

            # Accessing ticket options considered as client activity.
            ticket.touch()

            allow = ["OPTIONS"]

            if ticket.may("read"):
                allow.append("GET")
                options["features"] = BASE_FEATURES

            if ticket.may("write"):
                allow.extend(("PUT", "PATCH"))
                options["features"] = ALL_FEATURES

            # Backend specific options.
            options["max_readers"] = ctx.backend.max_readers
            options["max_writers"] = ctx.backend.max_writers

        resp.headers["allow"] = ",".join(allow)
        resp.send_json(options)



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import time
import socket

from contextlib import closing

from ovirt_imageio._internal import sockutil


def test_wait_for_unix_socket(tmpdir):
    addr = sockutil.UnixAddress(tmpdir.join("path"))

    # Socket was not created yet.
    start = time.monotonic()
    assert not sockutil.wait_for_socket(addr, 0.1)
    waited = time.monotonic() - start
    assert 0.1 <= waited < 0.2

    sock = socket.socket(socket.AF_UNIX)
    with closing(sock):
        sock.bind(addr)

        # Socket bound but not listening yet.
        start = time.monotonic()
        assert not sockutil.wait_for_socket(addr, 0.1)
        waited = time.monotonic() - start
        assert 0.1 <= waited < 0.2

        sock.listen(1)

        # Socket listening - should return immediately.
        assert sockutil.wait_for_socket(addr, 0.0)

    # Socket was closed - should return immediately.
    assert not sockutil.wait_for_socket(addr, 0.0)


def test_wait_for_tcp_socket():
    sock = socket.socket()
    with closing(sock):
        sock.bind(("localhost", 0))
        addr = sockutil.TCPAddress(*sock.getsockname())

        # Socket bound but not listening yet.
        start = time.monotonic()
        assert not sockutil.wait_for_socket(addr, 0.1)
        waited = time.monotonic() - start
        assert 0.1 <= waited < 0.2

        sock.listen(1)

        # Socket listening - should return immediately.
        assert sockutil.wait_for_socket(addr, 0.0)

    # Socket was closed - should return immediately.
    assert not sockutil.wait_for_socket(addr, 0.0)



# ovirt-imageio
# Copyright (C) 2015-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import http.client as http_client
import io
import json
import time

import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server

from .. import testutil
from .. import http

# Exceptions raised when peer closed the connection and auto_open is disabled.
# Raised in request() or in getresponse(), depends on timing.
CONNECTION_CLOSED = (
    http_client.NotConnected,
    http_client.BadStatusLine,
)


BASE_FEATURES = {"checksum", "extents"}
ALL_FEATURES = BASE_FEATURES | {"zero", "flush"}


@pytest.fixture(scope="module")
def srv():
    cfg = config.load(["test/conf/daemon.conf"])
    s = server.Server(cfg)
    s.start()
    yield s
    s.stop()


@pytest.fixture
def client(srv):
    with http.RemoteClient(srv.config) as c:
        yield c


def test_no_resource(srv, client):
    res = client.request("PUT", "/no/such/resource")
    assert res.status == 404


def test_no_method(srv, client):
    res = client.request("FOO", "/images/")
    assert res.status == 405


def test_upload_no_ticket_id(tmpdir, srv, client):
    res = client.put("/images/", "content")
    assert res.status == 400


def test_upload_no_ticket(tmpdir, srv, client):
    res = client.put("/images/no-such-ticket", "content")
    assert res.status == 403


def test_upload_forbidden(tmpdir, srv, client):
    ticket = testutil.create_ticket(
        url="file:///no/such/image", ops=["read"])
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "content")
    assert res.status == 403


def test_upload_content_length_missing(tmpdir, srv, client):
    ticket = testutil.create_ticket(url="file:///no/such/image")
    srv.auth.add(ticket)
    res = client.raw_request("PUT", "/images/" + ticket["uuid"])
    assert res.status == 400


def test_upload_content_length_invalid(tmpdir, srv, client):
    ticket = testutil.create_ticket(url="file:///no/such/image")
    srv.auth.add(ticket)
    res = client.raw_request("PUT", "/images/" + ticket["uuid"],
                             headers={"content-length": "invalid"})
    assert res.status == 400


def test_upload_content_length_negative(tmpdir, srv, client):
    image = testutil.create_tempfile(tmpdir, "image", b"before")
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    res = client.raw_request("PUT", "/images/" + ticket["uuid"],
                             headers={"content-length": "-1"})
    assert res.status == 400


def test_upload_no_content(tmpdir, srv, client):
    # This is a pointless request, but valid
    image = testutil.create_tempfile(tmpdir, "image", b"before")
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "")
    assert res.status == 200


def test_upload_extends_ticket(tmpdir, srv, client, fake_time):
    image = testutil.create_tempfile(tmpdir, "image", b"before")
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 300

    fake_time.now += 200
    res = client.put("/images/" + ticket["uuid"], "")
    assert res.status == 200

    res.read()

    # Yield to server thread - will close the opreration and extend the
    # ticket.
    time.sleep(0.1)

    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 500


# TODO: test that flush actually flushes data. Current tests just verify that
# the server does not reject the query string.
@pytest.mark.parametrize("flush", [None, "y", "n"])
def test_upload(tmpdir, srv, client, flush):
    data = b"-------|after"
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    uri = "/images/" + ticket["uuid"]
    if flush:
        uri += "?flush=" + flush
    res = client.put(uri, "content")
    with io.open(str(image), "rb") as f:
        assert f.read(len(data)) == b"content|after"
    assert res.status == 200
    assert res.getheader("content-length") == "0"


def test_upload_invalid_flush(tmpdir, srv, client):
    ticket = testutil.create_ticket(url="file:///no/such/image")
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"] + "?flush=invalid", "data")
    assert res.status == 400


@pytest.mark.parametrize("crange,before,after", [
    ("bytes 7-13/20", b"before|-------|after", b"before|content|after"),
    ("bytes 0-6/*", b"-------|after", b"content|after"),
    ("bytes 0-*/*", b"-------|after", b"content|after"),
])
def test_upload_with_range(tmpdir, srv, client, crange, before, after):
    image = testutil.create_tempfile(tmpdir, "image", before)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "content",
                     headers={"Content-Range": crange})
    with io.open(str(image), "rb") as f:
        assert f.read(len(after)) == after
    assert res.status == 200


def test_upload_max_size(tmpdir, srv, client):
    image_size = 100
    content = b"b" * image_size
    image = testutil.create_tempfile(tmpdir, "image", b"")
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=image_size)
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], content)
    assert res.status == 200
    with io.open(str(image), "rb") as f:
        assert f.read(len(content)) == content


def test_upload_too_big(tmpdir, srv, client):
    image_size = 100
    image = testutil.create_tempfile(tmpdir, "image", b"")
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=image_size)
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "b" * (image_size + 1))
    assert res.status == 416
    assert image.read() == ""


def test_upload_last_byte(tmpdir, srv, client):
    image_size = 100
    image = testutil.create_tempfile(tmpdir, "image", b"a" * image_size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=image_size)
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "b",
                     headers={"Content-Range": "bytes 99-100/*"})
    assert res.status == 200
    with io.open(str(image), "rb") as f:
        assert f.read(image_size) == b"a" * 99 + b"b"


def test_upload_after_last_byte(tmpdir, srv, client):
    image_size = 100
    image = testutil.create_tempfile(tmpdir, "image", b"a" * image_size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=image_size)
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "b",
                     headers={"Content-Range": "bytes 100-101/*"})
    assert res.status == 416
    assert image.read() == "a" * image_size


@pytest.mark.parametrize("content_range", [
    "",
    "   ",
    "7-13/20",
    "bytes invalid-invalid/*",
    "bytes 7-13/invalid",
    "bytes 7-13",
    "bytes 13-7/20",
])
def test_upload_invalid_range(tmpdir, srv, client, content_range):
    ticket = testutil.create_ticket()
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "content",
                     headers={"Content-Range": content_range})
    assert res.status == 400


def test_upload_close_connection(tmpdir, srv, client):
    image_size = 4096
    image = testutil.create_tempfile(tmpdir, "image", b"a" * image_size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=image_size)
    srv.auth.add(ticket)
    uri = "/images/" + ticket["uuid"] + "?close=y"
    data = b"b" * image_size

    # Disabling auto_open so we can test if a connection was closed.
    client.con.auto_open = False
    client.con.connect()

    res = client.put(uri, data)
    res.read()

    # Image should be updated...
    assert res.status == 200
    with io.open(str(image), "rb") as f:
        assert f.read() == data

    # But connection should be closed.
    assert res.getheader("connection") == "close"
    with pytest.raises(CONNECTION_CLOSED):
        client.put(uri, data)


def test_download_no_ticket_id(srv, client):
    res = client.get("/images/")
    assert res.status == http_client.BAD_REQUEST


def test_download_no_ticket(srv, client):
    res = client.get("/images/no-such-ticket")
    assert res.status == http_client.FORBIDDEN


@pytest.mark.parametrize("rng,start,end", [
    ("bytes=0-1023", 0, 1024),
    ("bytes=1-1023", 1, 1024),
    ("bytes=512-1023", 512, 1024),
    ("bytes=513-1023", 513, 1024),
    ("bytes=0-511", 0, 512),
    ("bytes=0-512", 0, 513),
])
def test_download(tmpdir, srv, client, rng, start, end):
    data = b"a" * 512 + b"b" * 512 + b"c" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=len(data))
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"], headers={"Range": rng})
    assert res.status == 206
    received = res.read()
    assert received == data[start:end]
    content_range = 'bytes %d-%d/%d' % (start, end-1, len(data))
    assert res.getheader("Content-Range") == content_range


def test_download_image_size_gt_ticket_size(tmpdir, srv, client):
    image = testutil.create_tempfile(tmpdir, "image", size=8192)
    ticket = testutil.create_ticket(url="file://" + str(image), size=4096)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"])
    assert res.status == 200
    assert len(res.read()) == 4096


def test_download_ticket_size_gt_image_size(tmpdir, srv, client):
    image = testutil.create_tempfile(tmpdir, "image", size=4096)
    ticket = testutil.create_ticket(url="file://" + str(image), size=8192)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"])
    assert res.status == 200
    assert len(res.read()) == 4096


def test_download_range_forbidden(tmpdir, srv, client):
    image = testutil.create_tempfile(tmpdir, "image", size=4096)
    ticket = testutil.create_ticket(url="file://" + str(image), size=8192)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"],
                     headers={"Range": "bytes=0-8192"})
    assert res.status == 416
    assert res.getheader("Content-Range") == "bytes */8192"


def test_download_range_unavailable(tmpdir, srv, client):
    image = testutil.create_tempfile(tmpdir, "image", size=8192)
    ticket = testutil.create_ticket(url="file://" + str(image), size=4096)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"],
                     headers={"Range": "bytes=0-4096"})
    assert res.status == 416
    assert res.getheader("Content-Range") == "bytes */4096"


def test_download_no_range(tmpdir, srv, client):
    size = 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"])
    assert res.status == 200
    received = res.read()
    assert received == b"\0" * size


def test_download_extends_ticket(tmpdir, srv, client, fake_time):
    size = 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size)
    srv.auth.add(ticket)
    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 300

    fake_time.now += 200
    res = client.get("/images/" + ticket["uuid"])
    assert res.status == 200

    res.read()

    # Yield to server thread - will close the opreration and extend the
    # ticket.
    time.sleep(0.1)

    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 500


def test_download_empty(tmpdir, srv, client):
    # Stupid edge case, but it should work, returning empty file :-)
    image = testutil.create_tempfile(tmpdir, "image")  # Empty image
    ticket = testutil.create_ticket(url="file://" + str(image), size=0)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"])
    assert res.status == 200
    data = res.read()
    assert data == b""


def test_download_partial_not_satistieble(tmpdir, srv, client):
    # Image is smaller than ticket size - may happen if engine failed to detect
    # actual image size reported by vdsm - one byte difference is enough to
    # cause a failure.
    # See https://bugzilla.redhat.com/1512315.
    size = 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size + 1)
    srv.auth.add(ticket)
    unsatisfiable_range = "bytes=0-%d" % size  # Max is size - 1
    res = client.get("/images/" + ticket["uuid"],
                     headers={"Range": unsatisfiable_range})
    assert res.status == http_client.REQUESTED_RANGE_NOT_SATISFIABLE


def test_download_partial_no_range(tmpdir, srv, client):
    # The image is smaller than the tiket size, but we don't request a range,
    # so we should get the existing length of the image, since the ticket size
    # is only an upper limit. Or maybe we should treat the ticket size as the
    # expected size?
    # This is another variant of https://bugzilla.redhat.com/1512315.
    size = 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size + 1)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"])
    assert res.status == http_client.OK
    # Should return the available image data, not the ticket size. Reading
    # this response will fail with IncompleteRead.
    assert res.length == 1024


def test_download_partial_no_range_empty(tmpdir, srv, client):
    # Image is empty, no range, should return an empty file - we return invalid
    # http response that fail on the client side with BadStatusLine: ''.
    # See https://bugzilla.redhat.com/1512312
    image = testutil.create_tempfile(tmpdir, "image")  # Empty image
    ticket = testutil.create_ticket(url="file://" + str(image), size=1024)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"])
    assert res.status == http_client.OK
    assert res.length == 0


def test_download_no_range_end(tmpdir, srv, client):
    size = 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"],
                     headers={"Range": "bytes=0-"})
    assert res.status == 206
    received = res.read()
    assert received == b"\0" * size


def test_download_holes(tmpdir, srv, client):
    size = 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"],
                     headers={"Range": "bytes=0-1023"})
    assert res.status == 206
    received = res.read()
    assert received == b"\0" * size


def test_download_filename_in_ticket(tmpdir, srv, client):
    size = 1024
    filename = "\u05d0.raw"  # hebrew aleph
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size,
                                    filename=filename)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"],
                     headers={"Range": "bytes=0-1023"})
    expected = "attachment; filename=\xd7\x90.raw"
    assert res.getheader("Content-Disposition") == expected


@pytest.mark.parametrize("rng,end", [
    ("bytes=0-1024", 512),
])
def test_download_out_of_range(tmpdir, srv, client, rng, end):
    data = b"a" * 512 + b"b" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image), size=end)
    srv.auth.add(ticket)
    res = client.get("/images/" + ticket["uuid"], headers={"Range": rng})
    res.read()
    assert res.status == 416


def test_download_progress(tmpdir, srv, client, monkeypatch):
    # We need to read at least one buffer to update the transfered value.
    monkeypatch.setattr(srv.config.backend_file, "buffer_size", 1024**2)

    # And we need to request enough data so the server does not complete before
    # the client read all the data.
    size = srv.config.backend_file.buffer_size * 50

    filename = tmpdir.join("image")
    with open(str(filename), 'wb') as image:
        image.truncate(size)
    ticket = testutil.create_ticket(
        url="file://" + str(filename), ops=["read"], size=size)
    srv.auth.add(ticket)
    ticket = srv.auth.get(ticket["uuid"])

    # No operations
    assert not ticket.active()
    assert ticket.transferred() == 0

    res = client.get("/images/" + ticket.uuid)
    res.read(srv.config.backend_file.buffer_size)

    # The server processed at least one buffer but we need to give it time
    # to touch the ticket.
    time.sleep(0.2)

    assert ticket.active()
    assert 0 < ticket.transferred() < size

    res.read()

    # The server has sent all the chunks but we need to give it time to
    # touch the ticket.
    time.sleep(0.2)

    assert not ticket.active()
    assert ticket.transferred() == size


def test_download_close_connection(tmpdir, srv, client):
    image_size = 4096
    image = testutil.create_tempfile(tmpdir, "image", b"a" * image_size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=image_size)
    srv.auth.add(ticket)
    uri = "/images/" + ticket["uuid"] + "?close=y"

    # Disabling auto_open so we can test if a connection was closed.
    client.con.auto_open = False
    client.con.connect()

    res = client.get(uri)
    data = res.read()

    # We should get image data...
    assert res.status == 200
    with io.open(str(image), "rb") as f:
        assert f.read() == data

    # But connection should be closed.
    assert res.getheader("connection") == "close"
    with pytest.raises(CONNECTION_CLOSED):
        client.get(uri)


# PATCH

def test_patch_unkown_op(srv, client):
    body = json.dumps({"op": "unknown"}).encode("ascii")
    res = client.patch("/images/no-such-uuid", body)
    assert res.status == 400


@pytest.mark.parametrize("msg", [
    {"op": "zero", "size": 20},
    {"op": "zero", "size": 20, "offset": 10},
    {"op": "zero", "size": 20, "offset": 10, "flush": True},
    {"op": "zero", "size": 20, "offset": 10, "future": True},
])
def test_zero(tmpdir, srv, client, msg):
    data = b"x" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    size = msg["size"]
    offset = msg.get("offset", 0)
    body = json.dumps(msg).encode("ascii")
    res = client.patch("/images/" + ticket["uuid"], body)

    assert res.status == 200
    assert res.getheader("content-length") == "0"
    with io.open(str(image), "rb") as f:
        assert f.read(offset) == data[:offset]
        assert f.read(size) == b"\0" * size
        assert f.read() == data[offset + size:]


def test_zero_extends_ticket(tmpdir, srv, client, fake_time):
    data = b"x" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 300

    fake_time.now += 200
    body = json.dumps({"op": "zero", "size": 512}).encode("ascii")
    res = client.patch("/images/" + ticket["uuid"], body)
    assert res.status == 200

    res.read()

    # Yield to server thread - will close the opreration and extend the
    # ticket.
    time.sleep(0.1)

    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 500


@pytest.mark.parametrize("msg", [
    {"op": "zero"},
    {"op": "zero", "size": "not an integer"},
    {"op": "zero", "size": -1},
    {"op": "zero", "size": 1, "offset": "not an integer"},
    {"op": "zero", "size": 1, "offset": -1},
    {"op": "zero", "size": 1, "offset": 1, "flush": "not a boolean"},
])
def test_zero_validation(srv, client, msg):
    body = json.dumps(msg).encode("ascii")
    res = client.patch("/images/no-such-uuid", body)
    assert res.status == 400


def test_zero_no_ticket_id(srv, client):
    body = json.dumps({"op": "zero", "size": 1}).encode("ascii")
    res = client.patch("/images/", body)
    assert res.status == 400


def test_zero_ticket_unknown(srv, client):
    body = json.dumps({"op": "zero", "size": 1}).encode("ascii")
    res = client.patch("/images/no-such-uuid", body)
    assert res.status == 403


def test_zero_ticket_readonly(tmpdir, srv, client):
    ticket = testutil.create_ticket(
        url="file:///no/such/image", ops=["read"])
    srv.auth.add(ticket)
    body = json.dumps({"op": "zero", "size": 1}).encode("ascii")
    res = client.patch("/images/" + ticket["uuid"], body)
    assert res.status == 403


# TODO: Test that data was flushed.
@pytest.mark.parametrize("msg", [
    {"op": "flush"},
    {"op": "flush", "future": True},
])
def test_flush(tmpdir, srv, client, msg):
    data = b"x" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    body = json.dumps(msg).encode("ascii")
    res = client.patch("/images/" + ticket["uuid"], body)

    assert res.status == 200
    assert res.getheader("content-length") == "0"


def test_flush_extends_ticket(tmpdir, srv, client, fake_time):
    data = b"x" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 300

    fake_time.now += 200
    body = json.dumps({"op": "flush"}).encode("ascii")
    res = client.patch("/images/" + ticket["uuid"], body)
    assert res.status == 200

    res.read()

    # Yield to server thread - will close the opreration and extend the
    # ticket.
    time.sleep(0.1)

    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 500


def test_flush_no_ticket_id(srv, client):
    body = json.dumps({"op": "flush"}).encode("ascii")
    res = client.patch("/images/", body)
    assert res.status == 400


def test_flush_ticket_unknown(srv, client):
    body = json.dumps({"op": "flush"}).encode("ascii")
    res = client.patch("/images/no-such-uuid", body)
    assert res.status == 403


def test_flush_ticket_readonly(tmpdir, srv, client):
    ticket = testutil.create_ticket(
        url="file:///no/such/image", ops=["read"])
    srv.auth.add(ticket)
    body = json.dumps({"op": "flush"}).encode("ascii")
    res = client.patch("/images/" + ticket["uuid"], body)
    assert res.status == 403


# Options

def test_options_all(srv, client):
    res = client.options("/images/*")
    allows = {"OPTIONS", "GET", "PUT", "PATCH"}
    assert res.status == 200
    assert set(res.getheader("allow").split(',')) == allows
    options = json.loads(res.read())
    assert set(options["features"]) == ALL_FEATURES
    assert options["unix_socket"] == srv.config.local.socket

    # Maximum connections reported only for actual ticket since this depends on
    # the backend.
    assert "max_readers" not in options
    assert "max_writers" not in options


def test_options_read_write(srv, client, tmpdir):
    size = 128 * 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=size, ops=["read", "write"])
    srv.auth.add(ticket)
    res = client.options("/images/" + ticket["uuid"])
    allows = {"OPTIONS", "GET", "PUT", "PATCH"}
    assert res.status == 200
    assert set(res.getheader("allow").split(',')) == allows
    options = json.loads(res.read())
    assert set(options["features"]) == ALL_FEATURES
    assert options["max_readers"] == srv.config.daemon.max_connections
    assert options["max_writers"] == 1  # Using file backend.


def test_options_read(srv, client, tmpdir):
    size = 128 * 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=size, ops=["read"])
    srv.auth.add(ticket)
    res = client.options("/images/" + ticket["uuid"])
    allows = {"OPTIONS", "GET"}
    assert res.status == 200
    assert set(res.getheader("allow").split(',')) == allows
    options = json.loads(res.read())
    assert set(options["features"]) == BASE_FEATURES
    assert options["max_readers"] == srv.config.daemon.max_connections
    assert options["max_writers"] == 1  # Using file backend.


def test_options_write(srv, client, tmpdir):
    size = 128 * 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=size, ops=["write"])
    srv.auth.add(ticket)
    res = client.options("/images/" + ticket["uuid"])
    # Having "write" imply also "read".
    allows = {"OPTIONS", "GET", "PUT", "PATCH"}
    assert res.status == 200
    assert set(res.getheader("allow").split(',')) == allows
    options = json.loads(res.read())
    assert set(options["features"]) == ALL_FEATURES
    assert options["max_readers"] == srv.config.daemon.max_connections
    assert options["max_writers"] == 1  # Using file backend.


def test_options_extends_ticket(srv, client, tmpdir, fake_time):
    size = 128 * 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size)
    srv.auth.add(ticket)
    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 300

    fake_time.now += 200
    res = client.options("/images/" + ticket["uuid"])
    assert res.status == 200

    res.read()

    # Yield to server thread - will close the opreration and extend the
    # ticket.
    time.sleep(0.1)

    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 500


def test_options_for_no_ticket(srv, client):
    res = client.options("/images/")
    assert res.status == 400


def test_options_for_nonexistent_ticket(srv, client):
    res = client.options("/images/no-such-ticket")
    assert res.status == 403


def test_options_ticket_expired(srv, client, tmpdir, fake_time):
    size = 128 * 1024
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=size, timeout=300)
    srv.auth.add(ticket)
    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 300

    # Make the ticket expire
    fake_time.now += 300
    res = client.options("/images/" + ticket["uuid"])
    assert res.status == 403

    server_ticket = srv.auth.get(ticket["uuid"]).info()
    assert server_ticket["expires"] == 300


# HTTP correctness

def test_response_version_success(tmpdir, srv, client):
    image = testutil.create_tempfile(tmpdir, "image", b"old")
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    res = client.put("/images/" + ticket["uuid"], "new")
    assert res.status == 200
    assert res.version == 11


def test_response_version_error(tmpdir, srv, client):
    res = client.get("/images/no-such-ticket")
    assert res.status != 200
    assert res.version == 11


@pytest.mark.parametrize("method, body", [
    ("PUT", "data"),
    ("PATCH", json.dumps({"op": "flush"}).encode("ascii")),
    ("OPTIONS", None),
    ("GET", None),
])
def test_keep_alive_connection_on_success(tmpdir, srv, client, method, body):
    # After successful request the connection should remain open.
    image = testutil.create_tempfile(tmpdir, "image", size=1024)
    ticket = testutil.create_ticket(url="file://" + str(image),
                                    size=1024)
    srv.auth.add(ticket)
    uri = "/images/%(uuid)s" % ticket
    # Disabling auto_open so we can test if a connection was closed.
    client.con.auto_open = False
    client.con.connect()

    # Send couple of requests - all should succeed.
    for i in range(3):
        r = client.request(method, uri, body=body)
        r.read()
        assert r.status == 200


@pytest.mark.parametrize("method, body", [
    ("OPTIONS", None),
    ("GET", None),
    # Patch reads entire body before checking ticket, so connection should be
    # kept alive.
    ("PATCH", json.dumps({"op": "flush"}).encode("ascii")),
    ("PUT", "data"),
])
def test_close_connection_on_auth_errors(tmpdir, srv, client, method, body):
    # When a request does not have a payload, the server can keep the
    # connection open after an error. However we close the connection after
    # authorization errors as part of cancellation support.
    uri = "/images/no-such-ticket"
    # Disabling auto_open so we can test if a connection was closed.
    client.con.auto_open = False
    client.con.connect()

    # Send the first request. It will fail before reading the
    # payload.
    r = client.request(method, uri, body=body)
    r.read()
    assert r.status == 403

    # Try to send another request. This will fail since the server closed
    # the connection, and we disabled auto_open.
    assert r.getheader("connection") == "close"
    with pytest.raises(CONNECTION_CLOSED):
        client.request(method, uri, body=body)


# CORS support


def test_cors_options_none(srv, client):
    res = client.options("/images/*")
    res.read()

    assert res.getheader("Access-Control-Allow-Origin") is None
    assert res.getheader("Access-Control-Allow-Headers") is None
    assert res.getheader("Access-Control-Allow-Methods") is None
    assert res.getheader("Access-Control-Max-Age") is None


def test_cors_options_some(srv, client):
    headers = {
        "Origin": "https://foo.example",
        "Access-Control-Request-Method": "PUT",
    }
    res = client.options("/images/*", headers=headers)
    res.read()

    assert res.getheader("Access-Control-Allow-Origin") == "*"
    assert res.getheader("Access-Control-Allow-Methods") == "OPTIONS,GET,PUT"
    assert res.getheader("Access-Control-Max-Age") == "86400"


def test_cors_options_all(srv, client):
    headers = {
        "Origin": "https://foo.example",
        "Access-Control-Request-Method": "PUT",
        "Access-Control-Request-Headers": "X-PINGOTHER, Content-Type",
    }
    res = client.options("/images/*", headers=headers)

    assert res.getheader("Access-Control-Allow-Origin") == "*"
    assert res.getheader("Access-Control-Allow-Headers") == "*"
    assert res.getheader("Access-Control-Allow-Methods") == "OPTIONS,GET,PUT"
    assert res.getheader("Access-Control-Max-Age") == "86400"


def test_cors_get_ok(tmpdir, srv, client):
    size = 8192
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size)
    srv.auth.add(ticket)
    uri = "/images/" + ticket["uuid"]
    headers = {"Origin": "http://foo.example"}

    res = client.get(uri, headers=headers)
    data = res.read()

    assert res.getheader("Access-Control-Allow-Origin") == "*"
    assert res.getheader("Access-Control-Max-Age") == "86400"
    assert res.getheader("Access-Control-Allow-Headers") is None
    assert res.getheader("Access-Control-Allow-Methods") is None
    assert res.status == 200
    assert data == b"\0" * size


def test_cors_get_error(srv, client):
    headers = {"Origin": "http://foo.example"}
    res = client.get("/images/no-such-ticket", headers=headers)
    res.read()

    assert res.getheader("Access-Control-Allow-Origin") == "*"
    assert res.getheader("Access-Control-Max-Age") == "86400"
    assert res.status == 403


def test_cors_put_ok(tmpdir, srv, client):
    size = 8192
    image = testutil.create_tempfile(tmpdir, "image", size=size)
    ticket = testutil.create_ticket(url="file://" + str(image), size=size)
    srv.auth.add(ticket)
    uri = "/images/" + ticket["uuid"]
    headers = {"Origin": "http://foo.example"}
    data = b"x" * size

    res = client.put(uri, body=data, headers=headers)
    res.read()

    assert res.getheader("Access-Control-Allow-Origin") == "*"
    assert res.getheader("Access-Control-Max-Age") == "86400"
    assert res.status == 200

    with open(image, "rb") as f:
        assert f.read() == data


def test_cors_put_error(srv, client):
    headers = {"Origin": "http://foo.example"}
    res = client.put("/images/no-such-ticket", body=b"x", headers=headers)
    res.read()

    assert res.getheader("Access-Control-Allow-Origin") == "*"
    assert res.getheader("Access-Control-Max-Age") == "86400"
    assert res.status == 403



# ovirt-imageio
# Copyright (C) 2017-2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
Test script for file.BlockIO() and file.FileIO(), simulating fallocate
and blkdiscard --zeroout.
"""

import argparse
import time
import urllib.parse

from ovirt_imageio._internal.backends import file


def humansize(s):
    if s.isdigit():
        return int(s)
    value, unit = s[:-1], s[-1]
    value = int(value)
    unit = unit.lower()
    if unit == "k":
        return value * 1024
    elif unit == "m":
        return value * 1024**2
    elif unit == "g":
        return value * 1024**3
    else:
        raise ValueError("Unsupported unit: %r" % unit)


parser = argparse.ArgumentParser()

parser.add_argument(
    "-l", "--length",
    dest="length",
    type=humansize,
    help=("The  number of bytes to zero (counting from the starting point) "
          "(default entire device or file)"))

parser.add_argument(
    "-o", "--offset",
    dest="offset",
    type=humansize,
    default=0,
    help="Byte offset into the device from which to start zeroing (default 0)")

parser.add_argument(
    "-p", "--step",
    dest="step",
    type=humansize,
    help=("The number of bytes to zero within one iteration. The default "
          "is to discard all by one ioctl call"))

parser.add_argument(
    "-s", "--sparse",
    dest="sparse",
    action="store_true",
    help="Deallocate zeroed space (punch holes)")

parser.add_argument(
    "filename",
    help="file or block device to fill with zeros")

args = parser.parse_args()

url = urllib.parse.urlparse("file://" + args.filename)

start_time = time.monotonic()

with file.open(url, "r+", sparse=args.sparse) as f:
    if args.length is None:
        args.length = f.size() - args.offset

    if args.step is None:
        args.step = args.length

    f.seek(args.offset)

    count = args.length
    while count:
        step = min(args.step, count)
        f.zero(step)
        count -= step

    f.flush()

elapsed_time = time.monotonic() - start_time

print("Zero %.2f GiB in %.3f seconds (%.2f GiB/s)" % (
    float(args.length) / 1024**3,
    elapsed_time,
    float(args.length) / 1024**3 / elapsed_time))



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from ovirt_imageio import client


class FakeTime:

    def __init__(self):
        self.now = 0.0

    def __call__(self):
        return self.now


class FakeFile:

    def __init__(self):
        self.last = None

    def write(self, s):
        self.last = s

    def flush(self):
        pass


def test_draw():
    fake_time = FakeTime()
    f = FakeFile()

    # Size is unknown at this point.
    pb = client.ProgressBar(output=f, step=0.1, now=fake_time)
    assert f.last == (
        "[ ------- ] 0 bytes, 0.00 seconds, 0 bytes/s".ljust(79) + "\r"
    )

    # Size was updated.
    fake_time.now += 0.1
    pb.size = 3 * 1024**3
    pb.update(0)
    assert f.last == (
        "[   0.00% ] 0 bytes, 0.10 seconds, 0 bytes/s".ljust(79) + "\r"
    )

    # Write some data...
    fake_time.now += 1.0
    pb.update(512 * 1024**2)
    assert f.last == (
        "[  16.67% ] 512.00 MiB, 1.10 seconds, 465.45 MiB/s".ljust(79) + "\r"
    )

    # Write zeros, much faster, but it is not time to update yet...
    fake_time.now += 0.05
    pb.update(512 * 1024**2)
    assert f.last == (
        "[  16.67% ] 512.00 MiB, 1.10 seconds, 465.45 MiB/s".ljust(79) + "\r"
    )

    # Write more zeors, time to update...
    fake_time.now += 0.05
    pb.update(512 * 1024**2)
    assert f.last == (
        "[  50.00% ] 1.50 GiB, 1.20 seconds, 1.25 GiB/s".ljust(79) + "\r"
    )

    # More zeros, rates increases...
    fake_time.now += 0.1
    pb.update(1024**3)
    assert f.last == (
        "[  83.33% ] 2.50 GiB, 1.30 seconds, 1.92 GiB/s".ljust(79) + "\r"
    )

    # More data, slow down again...
    fake_time.now += 1.0
    pb.update(512 * 1024**2)
    assert f.last == (
        "[ 100.00% ] 3.00 GiB, 2.30 seconds, 1.30 GiB/s".ljust(79) + "\r"
    )

    # Flush takes some time, lowering final rate.
    fake_time.now += 0.1
    pb.close()
    assert f.last == (
        "[ 100.00% ] 3.00 GiB, 2.40 seconds, 1.25 GiB/s".ljust(79) + "\n"
    )


def test_contextmanager():
    f = FakeFile()
    with client.ProgressBar(1024**3, output=f) as pb:
        pb.update(1024**3)
        assert f.last.endswith("\r")

    assert f.last.endswith("\n")



#!/usr/bin/python3
#
# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Configure Docker daemon. Has to be run with root privileges.
"""

import json
import subprocess

DAEMON_CONF = "/etc/docker/daemon.json"


def main():
    configure_docker()
    restart_docker_service()


def configure_ipv6(conf):
    """
    Add IPv6 support - enable IPv6 for docker and set IPv6 subnet for fixed
    IPs. The Docker setup is based on
    https://github.com/travis-ci/travis-ci/issues/8891#issuecomment-353403729
    """
    conf["ipv6"] = True
    conf["fixed-cidr-v6"] = "2001:1:1::/64"


def configure_docker():
    """
    Adjust exiting Docker daemon configuration. If there is none, create it.
    """
    try:
        with open(DAEMON_CONF, "r") as f:
            conf = json.loads(f.read())
    except FileNotFoundError:
        conf = {}

    configure_ipv6(conf)

    with open(DAEMON_CONF, "w") as f:
        f.write(json.dumps(conf))


def restart_docker_service():
    cmd = ["systemctl", "restart", "docker.service"]
    subprocess.run(
        cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


if __name__ == "__main__":
    main()



# ovirt-imageio
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

string = "2.4.1"

if __name__ == "__main__":
    print(string)
